! -*- f90 -*-
! Time-stamp: "2023-08-05 21:17:27 fujiwara"
!----------------------------------------------------------------------
!
!  EXFLIB (extended precision floating-point arithmetic library)
!  Fortran 90 Module
!
!  Copyright (C)  FUJIWARA Hiroshi  <fujiwara@acs.i.kyoto-u.ac.jp>
!
!----------------------------------------------------------------------
! TODO
!
! * Specifing precision in user program without changing this file
! * Functions (Bessel, Gamma, degree trigonometric, cubic)
! * type conversion by INT()
! * debug of output : WRITE(*,*) CMPLX( TYPE(ezfloat) )
! * rounding control, interval arithmetic
!----------------------------------------------------------------------

    MODULE exflib

      IMPLICIT NONE

#ifndef EXFLIB_EXFLOAT_PRECISION10

      !------------------------------------------------------------
      ! Your Requirement Digits (in Decimal)
      ! 40 <= exflib_exfloat_precision10 <= 19700
      !------------------------------------------------------------
      INTEGER*4, PARAMETER :: exflib_exfloat_precision10 = 1000

#else
      INTEGER*4, PARAMETER :: exflib_exfloat_precision10 = EXFLIB_EXFLOAT_PRECISION10
#endif

!**********************************************************************
! You do not need editing below this line.
!**********************************************************************

#ifndef EXFLIB_INTERFACE

#if (defined __GNUC__) || (defined __INTEL_COMPILER) || (defined __sun)
#define EXFLIB_INTERFACE 90
#endif

#if (defined __PGI)
#define EXFLIB_INTERFACE 77
#endif

#endif

      !------------------------------------------------------------
      ! Policy
      !
      ! * FORTRAN KEYWORDs are CAPITAL, others are small
      ! * Not allowed changing values in the module from other procedures.
      !   (Encapsulation, in the sense of OOP)
      !
      ! * Arguments order : (destination, src1, src2, src3, ...)
      ! * Inhibiting arithmetic with double, single
      ! * Allowing relation with double, single
      !------------------------------------------------------------
      ! Type Name (in I32LP64 data model)
      !
      ! x : exfloat   : multiple-precision number, real
      ! z : ezfloat   : multiple-precision number, complex
      ! q : quadword  : 64-bit width integer                (INTEGER*8)
      ! i : int       : 32-bit width integer                (INTEGER*4)
      !
      ! d : double    : 64-bit width floating point number  (REAL*8)
      ! s : single    : 32-bit width floating point number  (REAL*4)
      ! w : dcomplex  : double complex                      (COMPLEX*16)
      ! c : scomplex  : single complex                      (COMPLEX*8)
      !------------------------------------------------------------

      !------------------------------------------------------------
      ! Memory allocation size
      !
      ! 'd' digits precision can be stored in m_prec 64-bit integer elements,
      !   10**d = (2**64)**m_prec
      ! then
      !   m_prec  =  d * (log_2(10))/64  =  d / 19.265919722
      !------------------------------------------------------------
      INTEGER*8, PARAMETER, PRIVATE :: precision10 = exflib_exfloat_precision10
      INTEGER*8, PARAMETER, PRIVATE :: m_prec = (precision10 / 19.26591972) + 1

      !------------------------------------------------------------
      ! Actual Computation Precision, a size of 64-bit width integers.
      !
      ! 3 <= c_prec <= m_prec.
      !
      ! The lower 3 comes from implementation techniques.
      !------------------------------------------------------------
      INTEGER*8, PARAMETER, PRIVATE :: c_prec = m_prec

      !------------------------------------------------------------
      ! Public Member
      !------------------------------------------------------------
      INTEGER*4, PARAMETER :: exflib_exfloat_size = m_prec + 1
      INTEGER*4, PARAMETER :: exflib_exfloat_byte = (m_prec + 1)*8
      REAL*4,    PARAMETER :: exflib_exfloat_precision = c_prec * 19.26591972

      !------------------------------------------------------------
      ! max string length for formatted output
      !------------------------------------------------------------
      INTEGER*8, PARAMETER, PRIVATE :: max_string_length = precision10 * 2

      !------------------------------------------------------------
      ! working memory
      !------------------------------------------------------------
      INTEGER*8, PRIVATE, PARAMETER :: work_size = (1+m_prec+1)*5
      !INTEGER*8, PRIVATE :: mem(0:work_size)
      INTEGER*8, PRIVATE, PARAMETER :: funclog_invtbl_maxn = 500
      INTEGER*8, PRIVATE :: inv_table_log(0:(c_prec+2)*funclog_invtbl_maxn)
      INTEGER*8, PRIVATE :: inv_table_atan(0:(c_prec+2)*funclog_invtbl_maxn)

      ! body
      TYPE exfloat
         INTEGER*8 :: num(0:m_prec)
      END TYPE exfloat

      TYPE ezfloat
         TYPE(exfloat) :: re, im
      END TYPE ezfloat

      ! name and type, assembly or C routines
      ! (prototype declarations in the language C)

#if (EXFLIB_INTERFACE == 77)
#define exflib_float_add_FRT exflib_float_add_f77
#define exflib_float_sub_FRT exflib_float_sub_f77
#define exflib_float_mul_FRT exflib_float_mul_f77
#define exflib_float_div_FRT exflib_float_div_f77
#define exflib_float_imul_FRT exflib_float_imul_f77
#define exflib_float_idiv_FRT exflib_float_idiv_f77
#define exflib_float_lt_FRT exflib_float_lt_f77
#define exflib_float_round_near_FRT exflib_float_round_near_f77
#define exflib_float_round_chop_FRT exflib_float_round_chop_f77
#define exflib_float_round_plus_FRT exflib_float_round_plus_f77
#define exflib_float_round_minus_FRT exflib_float_round_minus_f77
#define exflib_float_etoef_FRT exflib_float_etoef_f77
#define exflib_float_eftod_FRT exflib_float_eftod_f77
#define exflib_float_ltoef_FRT exflib_float_ltoef_f77
#define exflib_float_dtoef_FRT exflib_float_dtoef_f77
#define exflib_float_ctoef_FRT exflib_float_ctoef_f77
#define exflib_float_fortran_format_FRT exflib_float_fortran_format_f77
#define exflib_float_fabs_FRT exflib_float_fabs_f77
#define exflib_float_sqrt_FRT exflib_float_sqrt_f77
#define exflib_float_sin_FRT exflib_float_sin_f77
#define exflib_float_cos_FRT exflib_float_cos_f77
#define exflib_float_exp_FRT exflib_float_exp_f77
#define exflib_float_pow_FRT exflib_float_pow_f77
#define exflib_float_pow_long_FRT exflib_float_pow_long_f77
#define exflib_float_log_FRT exflib_float_log_f77
#define exflib_float_asin_FRT exflib_float_asin_f77
#define exflib_float_atan_FRT exflib_float_atan_f77
#define exflib_float_modf_FRT exflib_float_modf_f77
#define exflib_float_fmod_FRT exflib_float_fmod_f77
#define exflib_float_ceil_FRT exflib_float_ceil_f77
#define exflib_float_floor_FRT exflib_float_floor_f77
#define exflib_float_sign_FRT exflib_float_sign_f77
#define exflib_float_exponent_FRT exflib_float_exponent_f77
#define exflib_float_fraction_FRT exflib_float_fraction_f77
#define exflib_float_nearest_FRT exflib_float_nearest_f77
#define exflib_float_set_exponent_FRT exflib_float_set_exponent_f77
#define exflib_init_reciprocal_table_FRT exflib_init_reciprocal_table_f77
#endif

#if (EXFLIB_INTERFACE == 90)
#define exflib_float_add_FRT exflib_float_add_f90
#define exflib_float_sub_FRT exflib_float_sub_f90
#define exflib_float_mul_FRT exflib_float_mul_f90
#define exflib_float_div_FRT exflib_float_div_f90
#define exflib_float_imul_FRT exflib_float_imul_f90
#define exflib_float_idiv_FRT exflib_float_idiv_f90
#define exflib_float_lt_FRT exflib_float_lt_f90
#define exflib_float_round_near_FRT exflib_float_round_near_f90
#define exflib_float_round_chop_FRT exflib_float_round_chop_f90
#define exflib_float_round_plus_FRT exflib_float_round_plus_f90
#define exflib_float_round_minus_FRT exflib_float_round_minus_f90
#define exflib_float_etoef_FRT exflib_float_etoef_f90
#define exflib_float_eftod_FRT exflib_float_eftod_f90
#define exflib_float_ltoef_FRT exflib_float_ltoef_f90
#define exflib_float_dtoef_FRT exflib_float_dtoef_f90
#define exflib_float_ctoef_FRT exflib_float_ctoef_f90
#define exflib_float_fortran_format_FRT exflib_float_fortran_format_f90
#define exflib_float_fabs_FRT exflib_float_fabs_f90
#define exflib_float_sqrt_FRT exflib_float_sqrt_f90
#define exflib_float_sin_FRT exflib_float_sin_f90
#define exflib_float_cos_FRT exflib_float_cos_f90
#define exflib_float_exp_FRT exflib_float_exp_f90
#define exflib_float_pow_FRT exflib_float_pow_f90
#define exflib_float_pow_long_FRT exflib_float_pow_long_f90
#define exflib_float_log_FRT exflib_float_log_f90
#define exflib_float_asin_FRT exflib_float_asin_f90
#define exflib_float_atan_FRT exflib_float_atan_f90
#define exflib_float_modf_FRT exflib_float_modf_f90
#define exflib_float_fmod_FRT exflib_float_fmod_f90
#define exflib_float_ceil_FRT exflib_float_ceil_f90
#define exflib_float_floor_FRT exflib_float_floor_f90
#define exflib_float_sign_FRT exflib_float_sign_f90
#define exflib_float_exponent_FRT exflib_float_exponent_f90
#define exflib_float_fraction_FRT exflib_float_fraction_f90
#define exflib_float_nearest_FRT exflib_float_nearest_f90
#define exflib_float_set_exponent_FRT exflib_float_set_exponent_f90
#define exflib_init_reciprocal_table_FRT exflib_init_reciprocal_table_f90
#endif

      INTERFACE

         INTEGER*8 FUNCTION exflib_float_add_FRT(x, a, b, pr64)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: x
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: a, b
           INTEGER*8, INTENT(IN) :: pr64
         END FUNCTION exflib_float_add_FRT

         INTEGER*8 FUNCTION exflib_float_sub_FRT(x, a, b, pr64)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: x
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: a, b
           INTEGER*8, INTENT(IN) :: pr64
         END FUNCTION exflib_float_sub_FRT

         INTEGER*8 FUNCTION exflib_float_mul_FRT(x, a, b, pr64)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: x
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: a, b
           INTEGER*8, INTENT(IN) :: pr64
         END FUNCTION exflib_float_mul_FRT

         INTEGER*8 FUNCTION exflib_float_div_FRT(x, a, b, pr64)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: x
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: a, b
           INTEGER*8, INTENT(IN) :: pr64
         END FUNCTION exflib_float_div_FRT

         INTEGER*8 FUNCTION exflib_float_imul_FRT(x, a, b, pr64)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: x
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: a
           INTEGER*8, INTENT(IN) :: b
           INTEGER*8, INTENT(IN) :: pr64
         END FUNCTION exflib_float_imul_FRT

         INTEGER*8 FUNCTION exflib_float_idiv_FRT(x, a, b, pr64)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: x
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: a
           INTEGER*8, INTENT(IN) :: b
           INTEGER*8, INTENT(IN) :: pr64
         END FUNCTION exflib_float_idiv_FRT

         LOGICAL*8 FUNCTION exflib_float_lt_FRT(a, b, pr64)
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: a, b
           INTEGER*8, INTENT(IN) :: pr64
         END FUNCTION exflib_float_lt_FRT

         SUBROUTINE exflib_float_round_near_FRT(x, hint, pr64)
           INTEGER*8, DIMENSION(0:), INTENT(INOUT) :: x
           INTEGER*8, INTENT(IN) :: hint, pr64
         END SUBROUTINE exflib_float_round_near_FRT

         SUBROUTINE exflib_float_round_chop_FRT(x, hint, pr64)
           INTEGER*8, DIMENSION(0:), INTENT(INOUT) :: x
           INTEGER*8, INTENT(IN) :: hint, pr64
         END SUBROUTINE exflib_float_round_chop_FRT

         SUBROUTINE exflib_float_round_plus_FRT(x, hint, pr64)
           INTEGER*8, DIMENSION(0:), INTENT(INOUT) :: x
           INTEGER*8, INTENT(IN) :: hint, pr64
         END SUBROUTINE exflib_float_round_plus_FRT

         SUBROUTINE exflib_float_round_minus_FRT(x, hint, pr64)
           INTEGER*8, DIMENSION(0:), INTENT(INOUT) :: x
           INTEGER*8, INTENT(IN) :: hint, pr64
         END SUBROUTINE exflib_float_round_minus_FRT

         SUBROUTINE exflib_float_etoef_FRT(dst_exfloat, src_str, pr64)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: dst_exfloat
           CHARACTER(*), INTENT(IN) :: src_str
           INTEGER*8, INTENT(IN) :: pr64
         END SUBROUTINE exflib_float_etoef_FRT

         REAL*8 FUNCTION exflib_float_eftod_FRT(src_exfloat)
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: src_exfloat
         END FUNCTION exflib_float_eftod_FRT

         SUBROUTINE exflib_float_ltoef_FRT(dst_exfloat, src_long, pr64)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: dst_exfloat
           INTEGER*8, INTENT(IN) :: src_long, pr64
         END SUBROUTINE exflib_float_ltoef_FRT

         SUBROUTINE exflib_float_dtoef_FRT(dst_exfloat, src_double, pr64)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: dst_exfloat
           REAL*8, INTENT(IN) :: src_double
           INTEGER*8, INTENT(IN) :: pr64
         END SUBROUTINE exflib_float_dtoef_FRT

         SUBROUTINE exflib_float_ctoef_FRT(dst_exfloat, src_str, pr64)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: dst_exfloat
           CHARACTER(*), INTENT(IN) :: src_str
           INTEGER*8, INTENT(IN) :: pr64
         END SUBROUTINE exflib_float_ctoef_FRT

         SUBROUTINE exflib_float_fortran_format_FRT(dst_str, form_str, src_exfloat, pr64, dst_len)
           CHARACTER(*), INTENT(OUT) :: dst_str
           CHARACTER(*), INTENT(IN) :: form_str
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: src_exfloat
           INTEGER*8, INTENT(IN) :: pr64, dst_len
         END SUBROUTINE exflib_float_fortran_format_FRT

         SUBROUTINE exflib_float_fabs_FRT(dst_exfloat, src_exfloat, pr64)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: dst_exfloat
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: src_exfloat
           INTEGER*8, INTENT(IN) :: pr64
         END SUBROUTINE exflib_float_fabs_FRT

         INTEGER*8 FUNCTION exflib_float_sqrt_FRT(dst_exfloat, src_exfloat, pr64, work_space)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: dst_exfloat, work_space
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: src_exfloat
           INTEGER*8, INTENT(IN) :: pr64
         END FUNCTION exflib_float_sqrt_FRT

         INTEGER*8 FUNCTION exflib_float_sin_FRT(dst_exfloat, src_exfloat, pr64, work_space)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: dst_exfloat, work_space
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: src_exfloat
           INTEGER*8, INTENT(IN) :: pr64
         END FUNCTION exflib_float_sin_FRT

         INTEGER*8 FUNCTION exflib_float_cos_FRT(dst_exfloat, src_exfloat, pr64, work_space)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: dst_exfloat, work_space
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: src_exfloat
           INTEGER*8, INTENT(IN) :: pr64
         END FUNCTION exflib_float_cos_FRT

         INTEGER*8 FUNCTION exflib_float_exp_FRT(dst_exfloat, src_exfloat, pr64, work_space)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: dst_exfloat, work_space
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: src_exfloat
           INTEGER*8, INTENT(IN) :: pr64
         END FUNCTION exflib_float_exp_FRT

         INTEGER*8 FUNCTION exflib_float_pow_FRT(dst_exfloat, base_exfloat, pow_exfloat, pr64, work_space, table)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: dst_exfloat, work_space, table
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: base_exfloat, pow_exfloat
           INTEGER*8, INTENT(IN) :: pr64
         END FUNCTION exflib_float_pow_FRT

         INTEGER*8 FUNCTION exflib_float_pow_long_FRT(dst_exfloat, base_exfloat, pow_q, pr64, work_space)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: dst_exfloat, work_space
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: base_exfloat
           INTEGER*8, INTENT(IN) :: pow_q
           INTEGER*8, INTENT(IN) :: pr64
         END FUNCTION exflib_float_pow_long_FRT

         INTEGER*8 FUNCTION exflib_float_log_FRT(dst_exfloat, src_exfloat, pr64, work_space, table)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: dst_exfloat, work_space, table
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: src_exfloat
           INTEGER*8, INTENT(IN) :: pr64
         END FUNCTION exflib_float_log_FRT

         INTEGER*8 FUNCTION exflib_float_asin_FRT(dst_exfloat, src_exfloat, pr64, work_space)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: dst_exfloat, work_space
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: src_exfloat
           INTEGER*8, INTENT(IN) :: pr64
         END FUNCTION exflib_float_asin_FRT

         INTEGER*8 FUNCTION exflib_float_atan_FRT(dst_exfloat, src_exfloat, pr64, work_space, table)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: dst_exfloat, work_space, table
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: src_exfloat
           INTEGER*8, INTENT(IN) :: pr64
         END FUNCTION exflib_float_atan_FRT

         SUBROUTINE exflib_init_reciprocal_table_FRT(tbl, size)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: tbl
           INTEGER*8, INTENT(IN) :: size
         END SUBROUTINE exflib_init_reciprocal_table_FRT

         SUBROUTINE exflib_float_modf_FRT(src_exfloat, dst_ip, dst_fp, pr64)
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: src_exfloat
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: dst_ip, dst_fp
           INTEGER*8, INTENT(IN) :: pr64
         END SUBROUTINE exflib_float_modf_FRT

         SUBROUTINE exflib_float_fmod_FRT(dst_r, src_d, src_m, pr64, work_space)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: dst_r
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: src_d, src_m
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: work_space
           INTEGER*8, INTENT(IN) :: pr64
         END SUBROUTINE exflib_float_fmod_FRT

         SUBROUTINE exflib_float_ceil_FRT(dst_ip, src_exfloat, pr64, work_space)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: dst_ip
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: src_exfloat
           INTEGER*8, INTENT(IN) :: pr64
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: work_space
         END SUBROUTINE exflib_float_ceil_FRT

         SUBROUTINE exflib_float_floor_FRT(dst_ip, src_exfloat, pr64, work_space)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: dst_ip
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: src_exfloat
           INTEGER*8, INTENT(IN) :: pr64
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: work_space
         END SUBROUTINE exflib_float_floor_FRT

         SUBROUTINE exflib_float_sign_FRT(dst, src_exfloat, sign, pr64)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: dst
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: src_exfloat, sign
           INTEGER*8, INTENT(IN) :: pr64
         END SUBROUTINE exflib_float_sign_FRT

         SUBROUTINE exflib_float_exponent_FRT(dst, src_exfloat)
           INTEGER*8, INTENT(OUT) :: dst
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: src_exfloat
         END SUBROUTINE exflib_float_exponent_FRT

         SUBROUTINE exflib_float_fraction_FRT(dst_ip, src_exfloat, pr64)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: dst_ip
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: src_exfloat
           INTEGER*8, INTENT(IN) :: pr64
         END SUBROUTINE exflib_float_fraction_FRT

         SUBROUTINE exflib_float_nearest_FRT(dst, src_exfloat, sign, pr64)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: dst
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: src_exfloat, sign
           INTEGER*8, INTENT(IN) :: pr64
         END SUBROUTINE exflib_float_nearest_FRT

         SUBROUTINE exflib_float_set_exponent_FRT(dst, src_exfloat, e, pr64)
           INTEGER*8, DIMENSION(0:), INTENT(OUT) :: dst
           INTEGER*8, DIMENSION(0:), INTENT(IN) :: src_exfloat
           INTEGER*8, INTENT(IN) :: e
           INTEGER*8, INTENT(IN) :: pr64
         END SUBROUTINE exflib_float_set_exponent_FRT

     END INTERFACE

      !------------------------------------------------------------
      ! interface definition
      !------------------------------------------------------------

      !----------------------------------------
      ! arithmetic
      !----------------------------------------
      INTERFACE OPERATOR ( + )
         MODULE PROCEDURE exflib_float_add_mm
         MODULE PROCEDURE exflib_float_add_mq
         MODULE PROCEDURE exflib_float_add_mi
         MODULE PROCEDURE exflib_float_add_qm
         MODULE PROCEDURE exflib_float_add_im
         MODULE PROCEDURE exflib_float_add_unarray

         MODULE PROCEDURE exflib_float_add_zz
         MODULE PROCEDURE exflib_float_add_zm
         MODULE PROCEDURE exflib_float_add_mz
         MODULE PROCEDURE exflib_float_add_zq
         MODULE PROCEDURE exflib_float_add_qz
         MODULE PROCEDURE exflib_float_add_zi
         MODULE PROCEDURE exflib_float_add_iz
         MODULE PROCEDURE exflib_float_add_zunarray
      END INTERFACE

      INTERFACE OPERATOR ( - )
         MODULE PROCEDURE exflib_float_sub_mm
         MODULE PROCEDURE exflib_float_sub_mq
         MODULE PROCEDURE exflib_float_sub_mi
         MODULE PROCEDURE exflib_float_sub_qm
         MODULE PROCEDURE exflib_float_sub_im
         MODULE PROCEDURE exflib_float_sub_unarray

         MODULE PROCEDURE exflib_float_sub_zz
         MODULE PROCEDURE exflib_float_sub_zm
         MODULE PROCEDURE exflib_float_sub_mz
         MODULE PROCEDURE exflib_float_sub_zq
         MODULE PROCEDURE exflib_float_sub_qz
         MODULE PROCEDURE exflib_float_sub_zi
         MODULE PROCEDURE exflib_float_sub_iz
         MODULE PROCEDURE exflib_float_sub_zunarray
      END INTERFACE

      INTERFACE OPERATOR ( * )
         MODULE PROCEDURE exflib_float_mul_mm
         MODULE PROCEDURE exflib_float_mul_mq
         MODULE PROCEDURE exflib_float_mul_mi
         MODULE PROCEDURE exflib_float_mul_qm
         MODULE PROCEDURE exflib_float_mul_im

         MODULE PROCEDURE exflib_float_mul_zz
         MODULE PROCEDURE exflib_float_mul_zm
         MODULE PROCEDURE exflib_float_mul_mz
         MODULE PROCEDURE exflib_float_mul_zq
         MODULE PROCEDURE exflib_float_mul_qz
         MODULE PROCEDURE exflib_float_mul_zi
         MODULE PROCEDURE exflib_float_mul_iz
      END INTERFACE

      INTERFACE OPERATOR ( / )
         MODULE PROCEDURE exflib_float_div_mm
         MODULE PROCEDURE exflib_float_div_mq
         MODULE PROCEDURE exflib_float_div_mi
         MODULE PROCEDURE exflib_float_div_qm
         MODULE PROCEDURE exflib_float_div_im

         MODULE PROCEDURE exflib_float_div_zz
         MODULE PROCEDURE exflib_float_div_zm
         MODULE PROCEDURE exflib_float_div_mz
         MODULE PROCEDURE exflib_float_div_zq
         MODULE PROCEDURE exflib_float_div_qz
         MODULE PROCEDURE exflib_float_div_zi
         MODULE PROCEDURE exflib_float_div_iz
      END INTERFACE

      !----------------------------------------
      ! relation
      !----------------------------------------
      INTERFACE OPERATOR ( == )
         MODULE PROCEDURE exflib_float_eq_mm
         MODULE PROCEDURE exflib_float_eq_mq
         MODULE PROCEDURE exflib_float_eq_mi
         MODULE PROCEDURE exflib_float_eq_md
         MODULE PROCEDURE exflib_float_eq_ms
         MODULE PROCEDURE exflib_float_eq_qm
         MODULE PROCEDURE exflib_float_eq_im
         MODULE PROCEDURE exflib_float_eq_dm
         MODULE PROCEDURE exflib_float_eq_sm

         MODULE PROCEDURE exflib_float_eq_xw
         MODULE PROCEDURE exflib_float_eq_wx
         MODULE PROCEDURE exflib_float_eq_xc
         MODULE PROCEDURE exflib_float_eq_cx
         MODULE PROCEDURE exflib_float_eq_zz
         MODULE PROCEDURE exflib_float_eq_zw
         MODULE PROCEDURE exflib_float_eq_wz
         MODULE PROCEDURE exflib_float_eq_zc
         MODULE PROCEDURE exflib_float_eq_cz
         MODULE PROCEDURE exflib_float_eq_zm
         MODULE PROCEDURE exflib_float_eq_mz
         MODULE PROCEDURE exflib_float_eq_zq
         MODULE PROCEDURE exflib_float_eq_zi
         MODULE PROCEDURE exflib_float_eq_zd
         MODULE PROCEDURE exflib_float_eq_zs
         MODULE PROCEDURE exflib_float_eq_qz
         MODULE PROCEDURE exflib_float_eq_iz
         MODULE PROCEDURE exflib_float_eq_w
         MODULE PROCEDURE exflib_float_eq_c
      END INTERFACE

      INTERFACE OPERATOR ( /= )
         MODULE PROCEDURE exflib_float_ne_mm
         MODULE PROCEDURE exflib_float_ne_mq
         MODULE PROCEDURE exflib_float_ne_mi
         MODULE PROCEDURE exflib_float_ne_md
         MODULE PROCEDURE exflib_float_ne_ms
         MODULE PROCEDURE exflib_float_ne_qm
         MODULE PROCEDURE exflib_float_ne_im
         MODULE PROCEDURE exflib_float_ne_dm
         MODULE PROCEDURE exflib_float_ne_sm

         MODULE PROCEDURE exflib_float_ne_xw
         MODULE PROCEDURE exflib_float_ne_wx
         MODULE PROCEDURE exflib_float_ne_xc
         MODULE PROCEDURE exflib_float_ne_cx
         MODULE PROCEDURE exflib_float_ne_zz
         MODULE PROCEDURE exflib_float_ne_zw
         MODULE PROCEDURE exflib_float_ne_wz
         MODULE PROCEDURE exflib_float_ne_zc
         MODULE PROCEDURE exflib_float_ne_cz
         MODULE PROCEDURE exflib_float_ne_zm
         MODULE PROCEDURE exflib_float_ne_mz
         MODULE PROCEDURE exflib_float_ne_zq
         MODULE PROCEDURE exflib_float_ne_zi
         MODULE PROCEDURE exflib_float_ne_zd
         MODULE PROCEDURE exflib_float_ne_zs
         MODULE PROCEDURE exflib_float_ne_qz
         MODULE PROCEDURE exflib_float_ne_iz
         MODULE PROCEDURE exflib_float_ne_dz
         MODULE PROCEDURE exflib_float_ne_sz
      END INTERFACE

      INTERFACE OPERATOR ( < )
         MODULE PROCEDURE exflib_float_lt_mm
         MODULE PROCEDURE exflib_float_lt_mq
         MODULE PROCEDURE exflib_float_lt_mi
         MODULE PROCEDURE exflib_float_lt_md
         MODULE PROCEDURE exflib_float_lt_ms
         MODULE PROCEDURE exflib_float_lt_qm
         MODULE PROCEDURE exflib_float_lt_im
         MODULE PROCEDURE exflib_float_lt_dm
         MODULE PROCEDURE exflib_float_lt_sm
      END INTERFACE

      INTERFACE OPERATOR ( > )
         MODULE PROCEDURE exflib_float_gt_mm
         MODULE PROCEDURE exflib_float_gt_mq
         MODULE PROCEDURE exflib_float_gt_mi
         MODULE PROCEDURE exflib_float_gt_md
         MODULE PROCEDURE exflib_float_gt_ms
         MODULE PROCEDURE exflib_float_gt_qm
         MODULE PROCEDURE exflib_float_gt_im
         MODULE PROCEDURE exflib_float_gt_dm
         MODULE PROCEDURE exflib_float_gt_sm
      END INTERFACE

      INTERFACE OPERATOR ( >= )
         MODULE PROCEDURE exflib_float_ge_mm
         MODULE PROCEDURE exflib_float_ge_mq
         MODULE PROCEDURE exflib_float_ge_mi
         MODULE PROCEDURE exflib_float_ge_md
         MODULE PROCEDURE exflib_float_ge_ms
         MODULE PROCEDURE exflib_float_ge_qm
         MODULE PROCEDURE exflib_float_ge_im
         MODULE PROCEDURE exflib_float_ge_dm
         MODULE PROCEDURE exflib_float_ge_sm
      END INTERFACE

      INTERFACE OPERATOR ( <= )
         MODULE PROCEDURE exflib_float_le_mm
         MODULE PROCEDURE exflib_float_le_mq
         MODULE PROCEDURE exflib_float_le_mi
         MODULE PROCEDURE exflib_float_le_md
         MODULE PROCEDURE exflib_float_le_ms
         MODULE PROCEDURE exflib_float_le_qm
         MODULE PROCEDURE exflib_float_le_im
         MODULE PROCEDURE exflib_float_le_dm
         MODULE PROCEDURE exflib_float_le_sm
      END INTERFACE

      !----------------------------------------
      ! type conversion
      !----------------------------------------
      INTERFACE ASSIGNMENT ( = )
         MODULE PROCEDURE exflib_float_etoef_p
         MODULE PROCEDURE exflib_float_eftod_p
         MODULE PROCEDURE exflib_float_eftos_p
         MODULE PROCEDURE exflib_float_qtoef_p
         MODULE PROCEDURE exflib_float_ltoef_p

         MODULE PROCEDURE exflib_float_etoez
         MODULE PROCEDURE exflib_float_eftoez
         MODULE PROCEDURE exflib_float_qtoez
         MODULE PROCEDURE exflib_float_itoez
         MODULE PROCEDURE exflib_float_eztoc
         MODULE PROCEDURE exflib_float_eztow
      END INTERFACE

      INTERFACE exflib_cast
         MODULE PROCEDURE exflib_float_qqtoef_p
         MODULE PROCEDURE exflib_float_itoef_p
         MODULE PROCEDURE exflib_float_dtoef_p
         MODULE PROCEDURE exflib_float_stoef_p
         MODULE PROCEDURE exflib_float_wtoez_p
         MODULE PROCEDURE exflib_float_ctoez_p
!         MODULE PROCEDURE exflib_float_dtoez_p
!         MODULE PROCEDURE exflib_float_stoez_p
      END INTERFACE

      INTERFACE exflib_double
         MODULE PROCEDURE exflib_double_e
         MODULE PROCEDURE exflib_double_d
         MODULE PROCEDURE exflib_double_s
      END INTERFACE

      ! Return type is ambiguous, REAL*4 or REAL*8 ?
!      INTERFACE REAL
!         MODULE PROCEDURE exflib_double_e
!         MODULE PROCEDURE exflib_double_d
!         MODULE PROCEDURE exflib_double_s
!      END INTERFACE
      INTERFACE DBLE
         MODULE PROCEDURE exflib_double_e
         MODULE PROCEDURE exflib_double_d
         MODULE PROCEDURE exflib_double_s
      END INTERFACE

!      INTERFACE QREAL
!         MODULE PROCEDURE exflib_future_work
!      END INTERFACE

      !----------------------------------------
      ! auxiality : precision
      !----------------------------------------
      INTERFACE DIGITS
         MODULE PROCEDURE exflib_float_digits_m
      END INTERFACE
      INTERFACE RADIX
         MODULE PROCEDURE exflib_float_radix_m
      END INTERFACE
      INTERFACE PRECISION
         MODULE PROCEDURE exflib_float_precision_m
      END INTERFACE
      INTERFACE EPSILON
         MODULE PROCEDURE exflib_float_epsilon_m
      END INTERFACE
      INTERFACE HUGE
         MODULE PROCEDURE exflib_float_huge_m
      END INTERFACE
      INTERFACE TINY
         MODULE PROCEDURE exflib_float_tiny_m
      END INTERFACE
      INTERFACE MAXEXPONENT
         MODULE PROCEDURE exflib_float_maxexponent_m
      END INTERFACE
      INTERFACE MINEXPONENT
         MODULE PROCEDURE exflib_float_minexponent_m
      END INTERFACE
      INTERFACE RANGE
         MODULE PROCEDURE exflib_float_range_m
      END INTERFACE
      INTERFACE SIGN
         MODULE PROCEDURE exflib_float_sign_m
      END INTERFACE
      INTERFACE EXPONENT
         MODULE PROCEDURE exflib_float_exponent_m
      END INTERFACE
      INTERFACE FRACTION
         MODULE PROCEDURE exflib_float_fraction_m
      END INTERFACE
      INTERFACE NEAREST
         MODULE PROCEDURE exflib_float_nearest_m
      END INTERFACE
      INTERFACE SET_EXPONENT
         MODULE PROCEDURE exflib_float_set_exponent_mi
         MODULE PROCEDURE exflib_float_set_exponent_mq
      END INTERFACE
      INTERFACE SPACING
         MODULE PROCEDURE exflib_float_spacing_m
      END INTERFACE
      INTERFACE RRSPACING
         MODULE PROCEDURE exflib_float_rrspacing_m
      END INTERFACE

      !----------------------------------------
      ! complex type
      !----------------------------------------
      INTERFACE CMPLX
!         MODULE PROCEDURE exflib_cmplx_z
         MODULE PROCEDURE exflib_cmplx_e
         MODULE PROCEDURE exflib_cmplx_m
         MODULE PROCEDURE exflib_cmplx_q
         MODULE PROCEDURE exflib_cmplx_i

         MODULE PROCEDURE exflib_cmplx_ee
         MODULE PROCEDURE exflib_cmplx_em
         MODULE PROCEDURE exflib_cmplx_me
         MODULE PROCEDURE exflib_cmplx_eq
         MODULE PROCEDURE exflib_cmplx_qe
         MODULE PROCEDURE exflib_cmplx_ei
         MODULE PROCEDURE exflib_cmplx_ie

         MODULE PROCEDURE exflib_cmplx_mm
         MODULE PROCEDURE exflib_cmplx_mq
         MODULE PROCEDURE exflib_cmplx_mi
         MODULE PROCEDURE exflib_cmplx_qm
         MODULE PROCEDURE exflib_cmplx_im

         MODULE PROCEDURE exflib_cmplx_qq
         MODULE PROCEDURE exflib_cmplx_qi
         MODULE PROCEDURE exflib_cmplx_iq

         MODULE PROCEDURE exflib_cmplx_ii

         MODULE PROCEDURE exflib_cmplx_zw
      END INTERFACE

      INTERFACE exflib_complex
         MODULE PROCEDURE exflib_cmplx_zw
      END INTERFACE
      INTERFACE DCMPLX
         MODULE PROCEDURE exflib_cmplx_zw
      END INTERFACE
!      INTERFACE QCMPLX
!         MODULE PROCEDURE exflib_cmplx_future_work
!      END INTERFACE

      INTERFACE polar
         MODULE PROCEDURE exflib_polar_ee
         MODULE PROCEDURE exflib_polar_em
         MODULE PROCEDURE exflib_polar_me
         MODULE PROCEDURE exflib_polar_eq
         MODULE PROCEDURE exflib_polar_qe
         MODULE PROCEDURE exflib_polar_ei
         MODULE PROCEDURE exflib_polar_ie

         MODULE PROCEDURE exflib_polar_mm
         MODULE PROCEDURE exflib_polar_mq
         MODULE PROCEDURE exflib_polar_mi
         MODULE PROCEDURE exflib_polar_qm
         MODULE PROCEDURE exflib_polar_im

         MODULE PROCEDURE exflib_polar_qq
         MODULE PROCEDURE exflib_polar_qi
         MODULE PROCEDURE exflib_polar_iq

         MODULE PROCEDURE exflib_polar_ii
      END INTERFACE

      INTERFACE REAL
         MODULE PROCEDURE exflib_ezfloat_real
      END INTERFACE

!      INTERFACE imag
!         MODULE PROCEDURE exflib_ezfloat_imag
!      END INTERFACE
      INTERFACE AIMAG
         MODULE PROCEDURE exflib_ezfloat_aimag
      END INTERFACE

      INTERFACE CONJG
         MODULE PROCEDURE exflib_ezfloat_conjg
      END INTERFACE

      INTERFACE arg
         MODULE PROCEDURE exflib_ezfloat_arg
      END INTERFACE

      !--------------------------------------------------
      ! conversion to string with format (not completed)
      !--------------------------------------------------
      !
      ! prototype :
      !
      !     CHARACTER(*) FUNCTION exflib_format(format, exfloat_number)
      !       CHARACTER(*), INTENT(IN) :: format
      !       TYPE(exfloat), INTENT(IN) :: exfloat_number
      !
      ! examples :
      !
      !     CHARACTER(100) :: str
      !     TYPE(exfloat) :: x
      !     ...
      !     str = exflib_format('F10.8', x)
      !     str = exflib_format('E10.8', x)
      !     str = exflib_format('H', x)
      !     WRITE(*,*) TRIM(str)
      !     ...
      !     WRITE(*,*) TRIM(exflib_format('F10.8', x))
      !     WRITE(*,*) TRIM(exflib_format('E10.8', x))
      !     WRITE(*,*) TRIM(exflib_format('H', x))
      !
      ! format form: Fw.p
      !
      !     'F'   format specific character (default : 'H')
      !     'w'   total width (default : 6),
      !     'p'   precision, number of digits after the decimal point
      !           (default : 6)
      !
      !     Default values are defined in libexfloat.a (fortran_format.c)
      !
      ! format specific character:
      !
      !     (1) 'E'   scientific : exponential format
      !     (2) 'F'   fixed      : fixed point format
      !     (3) 'G'   general    : general format (not implemented)
      !     (4) 'H'   dump       : hexadecimal format
      !
      !     (1) and (2) are slow and have rounding error, because
      !     the type exfloat holds information in hexadecimal radix
      !     (more precisely, radix with 2**64) in memory.
      !     Mode (4) is fast and have no rounding error,
      !     however it is not familiear format with us...
      !
      ! error:
      !
      !     Format error is detected in run time.
      !     Unkown format specifiler is processed as hexadecimal format.
      !     In hexadecimal mode, w and p is ignored.
      !     If w < p, w is set to p.
      !
      INTERFACE exflib_format
         MODULE PROCEDURE exflib_format_f
         !MODULE PROCEDURE exflib_format_p
         MODULE PROCEDURE exflib_format_zf
      END INTERFACE

      !----------------------------------------
      ! mathematical function
      !----------------------------------------
      INTERFACE ABS
         MODULE PROCEDURE exflib_float_abs_f
         MODULE PROCEDURE exflib_float_abs_z
      END INTERFACE

      INTERFACE SQRT
         MODULE PROCEDURE exflib_float_sqrt_f
         MODULE PROCEDURE exflib_float_sqrt_z
      END INTERFACE

!      INTERFACE CUBIC
!         MODULE PROCEDURE exflib_float_cubic_f
!         MODULE PROCEDURE exflib_float_cubic_z
!      END INTERFACE

      INTERFACE SIN
         MODULE PROCEDURE exflib_float_sin_f
         MODULE PROCEDURE exflib_float_sin_z
      END INTERFACE
      INTERFACE COS
         MODULE PROCEDURE exflib_float_cos_f
         MODULE PROCEDURE exflib_float_cos_z
      END INTERFACE
      INTERFACE TAN
         MODULE PROCEDURE exflib_float_tan_f
         MODULE PROCEDURE exflib_float_tan_z
      END INTERFACE

      INTERFACE EXP
         MODULE PROCEDURE exflib_float_exp_f
         MODULE PROCEDURE exflib_float_exp_z
      END INTERFACE

      INTERFACE POW
         MODULE PROCEDURE exflib_float_pow_mm
         MODULE PROCEDURE exflib_float_pow_mq
         MODULE PROCEDURE exflib_float_pow_mi
         MODULE PROCEDURE exflib_float_pow_qm
         MODULE PROCEDURE exflib_float_pow_im

         MODULE PROCEDURE exflib_float_pow_zz
         MODULE PROCEDURE exflib_float_pow_zm
         MODULE PROCEDURE exflib_float_pow_mz
         MODULE PROCEDURE exflib_float_pow_zq
         MODULE PROCEDURE exflib_float_pow_zi
         MODULE PROCEDURE exflib_float_pow_qz
         MODULE PROCEDURE exflib_float_pow_iz
      END INTERFACE

      INTERFACE OPERATOR ( ** )
         MODULE PROCEDURE exflib_float_pow_mm
         MODULE PROCEDURE exflib_float_pow_mq
         MODULE PROCEDURE exflib_float_pow_mi
         MODULE PROCEDURE exflib_float_pow_qm
         MODULE PROCEDURE exflib_float_pow_im

         MODULE PROCEDURE exflib_float_pow_zz
         MODULE PROCEDURE exflib_float_pow_zm
         MODULE PROCEDURE exflib_float_pow_mz
         MODULE PROCEDURE exflib_float_pow_zq
         MODULE PROCEDURE exflib_float_pow_zi
         MODULE PROCEDURE exflib_float_pow_qz
         MODULE PROCEDURE exflib_float_pow_iz
      END INTERFACE

      INTERFACE LOG
         MODULE PROCEDURE exflib_float_log_f
         MODULE PROCEDURE exflib_float_log_z
      END INTERFACE
      INTERFACE LOG10
         MODULE PROCEDURE exflib_float_log10_f
         MODULE PROCEDURE exflib_float_log10_z
      END INTERFACE

      INTERFACE ASIN
         MODULE PROCEDURE exflib_float_asin_f
      END INTERFACE
      INTERFACE ACOS
         MODULE PROCEDURE exflib_float_acos_f
      END INTERFACE
      INTERFACE ATAN
         MODULE PROCEDURE exflib_float_atan_f
      END INTERFACE

      INTERFACE SINH
         MODULE PROCEDURE exflib_float_sinh_f
         MODULE PROCEDURE exflib_float_sinh_z
      END INTERFACE
      INTERFACE COSH
         MODULE PROCEDURE exflib_float_cosh_f
         MODULE PROCEDURE exflib_float_cosh_z
      END INTERFACE
      INTERFACE TANH
         MODULE PROCEDURE exflib_float_tanh_f
         MODULE PROCEDURE exflib_float_tanh_z
      END INTERFACE

      INTERFACE MOD
         MODULE PROCEDURE exflib_float_mod_f
      END INTERFACE
      INTERFACE INT
         MODULE PROCEDURE exflib_float_int_f
      END INTERFACE
      INTERFACE FLOOR
         MODULE PROCEDURE exflib_float_floor_f
      END INTERFACE
      INTERFACE CEILING
         MODULE PROCEDURE exflib_float_ceiling_f
      END INTERFACE

    CONTAINS

      !------------------------------------------------------------
      ! auxiality function
      !------------------------------------------------------------

      ! Getting actual computational precision (in decimal)
      !
      ! usage:
      !   write(*,*) 'Actual precision = ', exflib_get_precision()

      FUNCTION exflib_get_precision()
        REAL*4 exflib_get_precision
        exflib_get_precision = c_prec * 19.265919722
      END FUNCTION exflib_get_precision

      ! Getting required precision (in decimal)
      !
      ! usage:
      !   write(*,*) 'Required precision = ', exflib_get_req_precision()

      FUNCTION exflib_get_req_precision()
        INTEGER*8 exflib_get_req_precision
        exflib_get_req_precision = precision10
      END FUNCTION exflib_get_req_precision

      ! Getting byte size of a exfloat type number
      !
      ! usage:
      !    write(*,*) 'An exfloat is ', exflib_get_exfloat_byte(), ' bytes.'

      FUNCTION exflib_get_exfloat_byte()
        INTEGER*8 exflib_get_exfloat_byte
        exflib_get_exfloat_byte = (m_prec + 1)* 8
      END FUNCTION exflib_get_exfloat_byte

      ! Getting element size of a exfloat type number
      !
      ! usave:
      !    write(*,*) 'An exfloat consists of ', exflib_get_exfloat_elem(), ' integers.'

      FUNCTION exflib_get_exfloat_elem()
        INTEGER*8 exflib_get_exfloat_elem
        exflib_get_exfloat_elem = m_prec
      END FUNCTION exflib_get_exfloat_elem

      ! Getting the date and time libexfloat.a was compiled.
      !
      ! usage :
      !   write(*,*) 'Current libexfloat.a is ', exflib_version()
      !   write(*,*) 'Current libexfloat.a is ', exflib_get_version()

      FUNCTION exflib_version() RESULT (message)
        CHARACTER(64) :: message
        EXTERNAL :: exflib_compiled_date
        message = ''
        CALL exflib_compiled_date(message)
      END FUNCTION exflib_version

      FUNCTION exflib_get_version() RESULT (message)
        CHARACTER(64) :: message
        EXTERNAL :: exflib_compiled_date
        message = ''
        CALL exflib_compiled_date(message)
      END FUNCTION exflib_get_version

      ! computational digits in decimal
      FUNCTION exflib_float_digits_m(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8 :: x
        x = CEILING(c_prec * 19.265919722)
      END FUNCTION exflib_float_digits_m

      ! model radix
      FUNCTION exflib_float_radix_m(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8 :: x
        x = 2
      END FUNCTION exflib_float_radix_m

      ! computational precision in bits, including the hidden one
      FUNCTION exflib_float_precision_m(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8 :: x
        x = 1 + m_prec * 64
      END FUNCTION exflib_float_precision_m

      ! machine epsilon
      ! (the gap between 1 and the next floating-point value)
      ! (not the half of the gap)
      FUNCTION exflib_float_epsilon_m(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
        INTEGER*8 :: i
#if defined(__GNUC__) && (__GNUC__ >= 13)
        INTEGER*8, PARAMETER::EXP_BIAS = INT(Z'3fffffffffffffff', KIND=SELECTED_INT_KIND(16))
        x%num(0) = EXP_BIAS
#else
        x%num(0) = Z'3fffffffffffffff'
#endif
        x%num(0) = x%num(0) - (m_prec * 64)
        DO i = 1, c_prec
           x%num(i) = 0
        END DO

      END FUNCTION exflib_float_epsilon_m

      ! maximum number in exfloat
      FUNCTION exflib_float_huge_m(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
        INTEGER*8 :: i
#if defined(__GNUC__) && (__GNUC__ >= 13)
        INTEGER*8, PARAMETER::EXP_MAX = INT(Z'7ffffffffffffffe', KIND=SELECTED_INT_KIND(16))
        x%num(0) = EXP_MAX
#else
        x%num(0) = Z'7ffffffffffffffe'
#endif
        DO i = 1, c_prec
           x%num(i) = -1
        END DO
      END FUNCTION exflib_float_huge_m

      ! minimum number in exfloat
      FUNCTION exflib_float_tiny_m(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
        INTEGER*8 :: i
        ! normalized minimum number (subnormal is not supported)
        x%num(0) = 1
        DO i = 1, c_prec
           x%num(i) = 0
        END DO
      END FUNCTION exflib_float_tiny_m

      FUNCTION exflib_float_maxexponent_m(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8 :: x
#if defined(__GNUC__) && (__GNUC__ >= 13)
        INTEGER*8, PARAMETER::EXP_BIAS_P1 = INT(Z'4000000000000000', KIND=SELECTED_INT_KIND(16))
        x = EXP_BIAS_P1
#else
        x = Z'4000000000000000'
#endif
        ! convert to 'FORTRAN REAL model'
        x = x + 1
      END FUNCTION exflib_float_maxexponent_m

      FUNCTION exflib_float_minexponent_m(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8 :: x
#if defined(__GNUC__) && (__GNUC__ >= 13)
        INTEGER*8, PARAMETER::EXP_BIAS_M1 = INT(Z'3ffffffffffffffe', KIND=SELECTED_INT_KIND(16))
        x = EXP_BIAS_M1
#else
        x = Z'3ffffffffffffffe'
#endif
        x = -x
        ! convert to 'FORTRAN REAL model'
        x = x + 1
      END FUNCTION exflib_float_minexponent_m

      FUNCTION exflib_float_range_m(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8 :: x
        ! 0x3ffffffffffffffe = 4611686018427387902
        ! log10(2) * 0x3ffffffffffffffe = 1388255822130839282.468
#if defined(__GNUC__) && (__GNUC__ >= 13)
        INTEGER*8, PARAMETER::RANGE64 = INT(Z'134413509f79fef2', KIND=SELECTED_INT_KIND(16))
        x = RANGE64
#else
        x = Z'134413509f79fef2'
#endif
      END FUNCTION exflib_float_range_m

      FUNCTION exflib_float_sign_m(a, s) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a, s
        TYPE(exfloat) :: x
	INTEGER*8 :: prec
	prec = c_prec
        CALL exflib_float_sign_FRT(x%num, a%num, s%num, prec)
      END FUNCTION exflib_float_sign_m

      FUNCTION exflib_float_exponent_m(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8 :: x
        CALL exflib_float_exponent_FRT(x, a%num)
      END FUNCTION exflib_float_exponent_m

      FUNCTION exflib_float_fraction_m(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
	INTEGER*8 :: prec
	prec = c_prec
        CALL exflib_float_fraction_FRT(x%num, a%num, prec)
      END FUNCTION exflib_float_fraction_m

      FUNCTION exflib_float_nearest_m(a, s) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a, s
        TYPE(exfloat) :: x
	INTEGER*8 :: prec
	prec = c_prec
        CALL exflib_float_nearest_FRT(x%num, a%num, s%num, prec)
      END FUNCTION exflib_float_nearest_m

      FUNCTION exflib_float_set_exponent_mi(a, i) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: i
        TYPE(exfloat) :: x
        INTEGER*8 :: e
	INTEGER*8 :: prec
	prec = c_prec
        e = i
        CALL exflib_float_set_exponent_FRT(x%num, a%num, e, prec)
      END FUNCTION exflib_float_set_exponent_mi

      FUNCTION exflib_float_set_exponent_mq(a, e) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: e
        TYPE(exfloat) :: x
        INTEGER*8 :: prec
        prec = c_prec
        CALL exflib_float_set_exponent_FRT(x%num, a%num, e, prec)
      END FUNCTION exflib_float_set_exponent_mq

      FUNCTION exflib_float_spacing_m(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
        TYPE(exfloat) :: y
	INTEGER*8 :: prec
	prec = c_prec
        x = 1
        y = NEAREST(a,x)
        x = y - a
      END FUNCTION exflib_float_spacing_m

      FUNCTION exflib_float_rrspacing_m(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
        TYPE(exfloat) :: y
	INTEGER*8 :: prec
	prec = c_prec
        x = 1
        y = NEAREST(a,x)
        x = x / (y - a)
      END FUNCTION exflib_float_rrspacing_m

      !------------------------------------------------------------
      ! addition
      !------------------------------------------------------------
      ! exfloat + exfloat
      FUNCTION exflib_float_add_mm(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a, b
        TYPE(exfloat) :: x
        INTEGER*8 :: hint, prec
        prec = c_prec
        hint = exflib_float_add_FRT(x%num, a%num, b%num, prec)
        prec = c_prec
	CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_add_mm

      ! exfloat + quadword
      FUNCTION exflib_float_add_mq(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(exfloat) :: x
        INTEGER*8 :: hint, prec
        prec = c_prec
        CALL exflib_float_ltoef_FRT(x%num, b, prec)
        prec = c_prec
        hint = exflib_float_add_FRT(x%num, a%num, x%num, prec)
        prec = c_prec
	CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_add_mq

      ! exfloat + int
      FUNCTION exflib_float_add_mi(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(exfloat) :: x
        INTEGER*8 :: hint, prec, bb
        bb = b
        prec = c_prec
        CALL exflib_float_ltoef_FRT(x%num, bb, prec)
        prec = c_prec
        hint = exflib_float_add_FRT(x%num, a%num, x%num, prec)
        prec = c_prec
	CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_add_mi

      ! quadword + exfloat
      FUNCTION exflib_float_add_qm(a, b) RESULT(x)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat) :: x
        INTEGER*8 :: hint, prec
        prec = c_prec
        CALL exflib_float_ltoef_FRT(x%num, a, prec)
        prec = c_prec
        hint = exflib_float_add_FRT(x%num, x%num, b%num, prec)
        prec = c_prec
	CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_add_qm

      ! int + exfloat
      FUNCTION exflib_float_add_im(a, b) RESULT(x)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat) :: x
        INTEGER*8 :: hint, prec, aa
        aa = a
        prec = c_prec
        CALL exflib_float_ltoef_FRT(x%num, aa, prec)
        prec = c_prec
        hint = exflib_float_add_FRT(x%num, x%num, b%num, prec)
        prec = c_prec
	CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_add_im

      ! unarray operator (+)
      FUNCTION exflib_float_add_unarray(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
        INTEGER*8 :: i
        DO i = 0, c_prec
           x%num(i) = a%num(i)
        END DO
      END FUNCTION exflib_float_add_unarray

      ! ezfloat + ezfloat
      FUNCTION exflib_float_add_zz(a, b) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a, b
        TYPE(ezfloat) :: z
        z%re = a%re + b%re
        z%im = a%im + b%im
      END FUNCTION exflib_float_add_zz

      ! ezfloat + exfloat
      FUNCTION exflib_float_add_zm(a, b) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a%re + b
        z%im = a%im
      END FUNCTION exflib_float_add_zm

      ! exfloat + ezfloat
      FUNCTION exflib_float_add_mz(a, b) RESULT(z)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a + b%re
        z%im =     b%im
      END FUNCTION exflib_float_add_mz

      ! ezfloat + quadword
      FUNCTION exflib_float_add_zq(a, b) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a%re + b
        z%im = a%im
      END FUNCTION exflib_float_add_zq

      ! quadword + ezfloat
      FUNCTION exflib_float_add_qz(a, b) RESULT(z)
        INTEGER*8, INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a + b%re
        z%im =     b%im
      END FUNCTION exflib_float_add_qz

      ! ezfloat + int
      FUNCTION exflib_float_add_zi(a, b) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a%re + b
        z%im = a%im
      END FUNCTION exflib_float_add_zi

      ! int + ezfloat
      FUNCTION exflib_float_add_iz(a, b) RESULT(z)
        INTEGER*4, INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a + b%re
        z%im =     b%im
      END FUNCTION exflib_float_add_iz

      ! unarray operator (+)
      FUNCTION exflib_float_add_zunarray(a) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat) :: z
        z%re = a%re
        z%im = a%im
      END FUNCTION exflib_float_add_zunarray

      !------------------------------------------------------------
      ! subtraction
      !------------------------------------------------------------
      ! exfloat - exfloat
      FUNCTION exflib_float_sub_mm(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a, b
        TYPE(exfloat) :: x
        INTEGER*8 :: hint, prec
        prec = c_prec
        hint = exflib_float_sub_FRT(x%num, a%num, b%num, prec)
        prec = c_prec
	CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_sub_mm

      ! exfloat - quadword
      FUNCTION exflib_float_sub_mq(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(exfloat) :: x
        INTEGER*8 :: hint, prec
        prec = c_prec
        CALL exflib_float_ltoef_FRT(x%num, b, prec)
        prec = c_prec
        hint = exflib_float_sub_FRT(x%num, a%num, x%num, prec)
        prec = c_prec
        CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_sub_mq

      ! exfloat - int
      FUNCTION exflib_float_sub_mi(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(exfloat) :: x
        INTEGER*8 :: hint, prec, bb
        bb = b
        prec = c_prec
        CALL exflib_float_ltoef_FRT(x%num, bb, prec)
        prec = c_prec
        hint = exflib_float_sub_FRT(x%num, a%num, x%num, prec)
        prec = c_prec
	CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_sub_mi

      ! quadword - exfloat
      FUNCTION exflib_float_sub_qm(a, b) RESULT(x)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat) :: x
        INTEGER*8 :: hint, prec
        prec = c_prec
        CALL exflib_float_ltoef_FRT(x%num, a, prec)
        prec = c_prec
        hint = exflib_float_sub_FRT(x%num, x%num, b%num, prec)
        prec = c_prec
        CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_sub_qm

      ! int - exfloat
      FUNCTION exflib_float_sub_im(a, b) RESULT(x)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat) :: x
        INTEGER*8 :: hint, prec, aa
        aa = a
        prec = c_prec
        CALL exflib_float_ltoef_FRT(x%num, aa, prec)
        prec = c_prec
        hint = exflib_float_sub_FRT(x%num, x%num, b%num, prec)
        prec = c_prec
	CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_sub_im

      ! unarray operator (-)
      FUNCTION exflib_float_sub_unarray(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
        INTEGER*8 :: i
#if defined(__GNUC__) && (__GNUC__ >= 13)
        INTEGER*8, PARAMETER::SIGN_MASK = int(Z'8000000000000000', KIND=SELECTED_INT_KIND(16))
        INTEGER*8 :: flag
        flag = SIGN_MASK
#elif defined(__GNUC__)
        INTEGER*8 :: flag
        flag = Z'7fffffffffffffff'
        flag = -1 - flag;
#else
        INTEGER*8, PARAMETER :: flag = Z'8000000000000000'
#endif
        ! adding 0x8000000000000000, sign flag
        ! x%num(0) = a%num(0) - 9223372036854775808
        x%num(0) = a%num(0) + flag
        DO i = 1, c_prec
           x%num(i) = a%num(i)
        END DO
      END FUNCTION exflib_float_sub_unarray

      ! ezfloat - ezfloat
      FUNCTION exflib_float_sub_zz(a, b) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a, b
        TYPE(ezfloat) :: z
        z%re = a%re - b%re
        z%im = a%im - b%im
      END FUNCTION exflib_float_sub_zz

      ! ezfloat - exfloat
      FUNCTION exflib_float_sub_zm(a, b) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a%re - b
        z%im = a%im
      END FUNCTION exflib_float_sub_zm

      ! exfloat - ezfloat
      FUNCTION exflib_float_sub_mz(a, b) RESULT(z)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a - b%re
        z%im =   - b%im
      END FUNCTION exflib_float_sub_mz

      ! ezfloat - quadword
      FUNCTION exflib_float_sub_zq(a, b) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a%re - b
        z%im = a%im
      END FUNCTION exflib_float_sub_zq

      ! quadword - ezfloat
      FUNCTION exflib_float_sub_qz(a, b) RESULT(z)
        INTEGER*8, INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a - b%re
        z%im =   - b%im
      END FUNCTION exflib_float_sub_qz

      ! ezfloat - int
      FUNCTION exflib_float_sub_zi(a, b) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a%re - b
        z%im = a%im
      END FUNCTION exflib_float_sub_zi

      ! int - ezfloat
      FUNCTION exflib_float_sub_iz(a, b) RESULT(z)
        INTEGER*4, INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a - b%re
        z%im =   - b%im
      END FUNCTION exflib_float_sub_iz

      ! unarray operator (-)
      FUNCTION exflib_float_sub_zunarray(a) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat) :: z
        z%re = - a%re
        z%im = - a%im
      END FUNCTION exflib_float_sub_zunarray

      !------------------------------------------------------------
      ! multiplication
      !------------------------------------------------------------
      ! exfloat * exfloat
      FUNCTION exflib_float_mul_mm(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a, b
        TYPE(exfloat) :: x
        INTEGER*8 :: hint, prec
        prec = c_prec
        hint = exflib_float_mul_FRT(x%num, a%num, b%num, prec)
	prec = c_prec
	CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_mul_mm

      ! exfloat * quadword
      FUNCTION exflib_float_mul_mq(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(exfloat) :: x
        INTEGER*8 :: hint, prec
        prec = c_prec
        hint = exflib_float_imul_FRT(x%num, a%num, b, prec)
	prec = c_prec
	CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_mul_mq

      ! exfloat * int
      FUNCTION exflib_float_mul_mi(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(exfloat) :: x
        INTEGER*8 :: hint, prec
        INTEGER*8 :: bb
        bb = b
        prec = c_prec
        hint = exflib_float_imul_FRT(x%num, a%num, bb, prec)
	prec = c_prec
	CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_mul_mi

      ! quadword * exfloat
      FUNCTION exflib_float_mul_qm(a, b) RESULT(x)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat) :: x
        INTEGER*8 :: hint, prec
        prec = c_prec
        hint = exflib_float_imul_FRT(x%num, b%num, a, prec)
	prec = c_prec
	CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_mul_qm

      ! int * exfloat
      FUNCTION exflib_float_mul_im(a, b) RESULT(x)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat) :: x
        INTEGER*8 :: hint, prec
        INTEGER*8 :: aa
        aa = a
        prec = c_prec
        hint = exflib_float_imul_FRT(x%num, b%num, aa, prec)
	prec = c_prec
	CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_mul_im

      ! ezfloat * ezfloat
      FUNCTION exflib_float_mul_zz(a, b) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a, b
        TYPE(ezfloat) :: z
        z%re = a%re * b%re - a%im * b%im
        z%im = a%im * b%re + a%re * b%im
      END FUNCTION exflib_float_mul_zz

      ! ezfloat * exfloat
      FUNCTION exflib_float_mul_zm(a, b) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a%re * b
        z%im = a%im * b
      END FUNCTION exflib_float_mul_zm

      ! exfloat * ezfloat
      FUNCTION exflib_float_mul_mz(a, b) RESULT(z)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a * b%re
        z%im = a * b%im
      END FUNCTION exflib_float_mul_mz

      ! ezfloat * quadword
      FUNCTION exflib_float_mul_zq(a, b) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a%re * b
        z%im = a%im * b
      END FUNCTION exflib_float_mul_zq

      ! quadword * ezfloat
      FUNCTION exflib_float_mul_qz(a, b) RESULT(z)
        INTEGER*8, INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a * b%re
        z%im = a * b%im
      END FUNCTION exflib_float_mul_qz

      ! ezfloat * int
      FUNCTION exflib_float_mul_zi(a, b) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a%re * b
        z%im = a%im * b
      END FUNCTION exflib_float_mul_zi

      ! int * ezfloat
      FUNCTION exflib_float_mul_iz(a, b) RESULT(z)
        INTEGER*4, INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a * b%re
        z%im = a * b%im
      END FUNCTION exflib_float_mul_iz

      !------------------------------------------------------------
      ! division
      !------------------------------------------------------------
      ! exfloat / exfloat
      FUNCTION exflib_float_div_mm(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a, b
        TYPE(exfloat) :: x
        INTEGER*8 :: hint, prec
        prec = c_prec
        hint = exflib_float_div_FRT(x%num, a%num, b%num, prec)
	prec = c_prec
	CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_div_mm

      ! exfloat / quadword
      FUNCTION exflib_float_div_mq(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(exfloat) :: x
        INTEGER*8 :: hint, prec
        prec = c_prec
        hint = exflib_float_idiv_FRT(x%num, a%num, b, prec)
	prec = c_prec
	CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_div_mq

      ! exfloat / int
      FUNCTION exflib_float_div_mi(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(exfloat) :: x
        INTEGER*8 :: hint, prec
        INTEGER*8 :: bb
        bb = b
        prec = c_prec
        hint = exflib_float_idiv_FRT(x%num, a%num, bb, prec)
	prec = c_prec
	CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_div_mi

      ! quadword / exfloat
      FUNCTION exflib_float_div_qm(a, b) RESULT(x)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat) :: x
        INTEGER*8 :: hint, prec
        prec = c_prec
        CALL exflib_float_ltoef_FRT(x%num, a, prec)
        prec = c_prec
        hint = exflib_float_div_FRT(x%num, x%num, b%num, prec)
        prec = c_prec
        CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_div_qm

      ! int / exfloat
      FUNCTION exflib_float_div_im(a, b) RESULT(x)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat) :: x
        INTEGER*8 :: hint, prec, aa
        aa = a
        prec = c_prec
        CALL exflib_float_ltoef_FRT(x%num, aa, prec)
        prec = c_prec
        hint = exflib_float_div_FRT(x%num, x%num, b%num, prec)
	prec = c_prec
	CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_div_im

      ! ezfloat / ezfloat
      FUNCTION exflib_float_div_zz(a, b) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a, b
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: bb
        bb = b%re * b%re + b%im * b%im
        bb = 1 / bb
        z%re = ( a%re * b%re + a%im * b%im ) * bb
        z%im = ( a%im * b%re - a%re * b%im ) * bb
      END FUNCTION exflib_float_div_zz

      ! ezfloat / exfloat
      FUNCTION exflib_float_div_zm(a, b) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: bb
        bb = 1 / b
        z%re = a%re * bb
        z%im = a%im * bb
      END FUNCTION exflib_float_div_zm

      ! exfloat / ezfloat
      FUNCTION exflib_float_div_mz(a, b) RESULT(z)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: bb
        bb = b%re * b%re + b%im * b%im
        bb = 1 / bb
        z%re =   a * b%re * bb
        z%im = - a * b%im * bb
      END FUNCTION exflib_float_div_mz

      ! ezfloat / quadword
      FUNCTION exflib_float_div_zq(a, b) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a%re / b
        z%im = a%im / b
      END FUNCTION exflib_float_div_zq

      ! quadword / ezfloat
      FUNCTION exflib_float_div_qz(a, b) RESULT(z)
        INTEGER*8, INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: bb
        bb = b%re * b%re + b%im * b%im
        bb = 1 / bb
        z%re =   a * b%re * bb
        z%im = - a * b%im * bb
      END FUNCTION exflib_float_div_qz

      ! ezfloat / int
      FUNCTION exflib_float_div_zi(a, b) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a%re / b
        z%im = a%im / b
      END FUNCTION exflib_float_div_zi

      ! int / ezfloat
      FUNCTION exflib_float_div_iz(a, b) RESULT(z)
        INTEGER*4, INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: bb
        bb = b%re * b%re + b%im * b%im
        bb = 1 / bb
        z%re =   a * b%re * bb
        z%im = - a * b%im * bb
      END FUNCTION exflib_float_div_iz

      !------------------------------------------------------------
      ! type conversion
      !------------------------------------------------------------

      ! expression to exfloat (setting with decimal literal)
      SUBROUTINE exflib_float_etoef_p(x, a)
        TYPE(exfloat), INTENT(OUT) :: x
        CHARACTER(*), INTENT(IN) :: a
        INTEGER*8 :: prec = c_prec
        CALL exflib_float_etoef_FRT(x%num, a // ACHAR(0), prec)
      END SUBROUTINE exflib_float_etoef_p

      ! exfloat to double
      SUBROUTINE exflib_float_eftod_p(x, a)
        REAL*8, INTENT(OUT) :: x
        TYPE(exfloat), INTENT(IN) :: a
        x = exflib_float_eftod_FRT(a%num)
      END SUBROUTINE exflib_float_eftod_p

      ! exfloat to double
      FUNCTION exflib_double_e(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*8 :: x
        x = exflib_float_eftod_FRT(a%num)
      END FUNCTION exflib_double_e

      ! compatibility
      FUNCTION exflib_double_d(a) RESULT(x)
        REAL*8, INTENT(IN) :: a
        REAL*8 :: x
        x = a
      END FUNCTION exflib_double_d

      ! compatibility
      FUNCTION exflib_double_s(a) RESULT(x)
        REAL*4, INTENT(IN) :: a
        REAL*4 :: x
        x = a
      END FUNCTION exflib_double_s

      ! exfloat to single
      SUBROUTINE exflib_float_eftos_p(x, a)
        REAL*4, INTENT(OUT) :: x
        TYPE(exfloat), INTENT(IN) :: a
        REAL*8 :: xx
        xx = exflib_float_eftod_FRT(a%num)
        x = xx
      END SUBROUTINE exflib_float_eftos_p

      ! exfloat to single
      FUNCTION exflib_single(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*4 :: x
        REAL*8 :: xx
        xx = exflib_float_eftod_FRT(a%num)
        x = xx
      END FUNCTION exflib_single

      ! quadword to exfloat
      SUBROUTINE exflib_float_qtoef_p(x, a)
        TYPE(exfloat), INTENT(OUT) :: x
        INTEGER*8, INTENT(IN) :: a
        INTEGER*8 :: prec
        prec = c_prec
        CALL exflib_float_ltoef_FRT(x%num, a, prec)
      END SUBROUTINE exflib_float_qtoef_p

      ! int to exfloat
      SUBROUTINE exflib_float_ltoef_p(x, a)
        TYPE(exfloat), INTENT(OUT) :: x
        INTEGER*4, INTENT(IN) :: a
        INTEGER*8 :: prec, aa
        aa = a
        prec = c_prec
        CALL exflib_float_ltoef_FRT(x%num, aa, prec)
      END SUBROUTINE exflib_float_ltoef_p

      ! long to exfloat (function form)
      FUNCTION exflib_float_qqtoef_p(a) RESULT(x)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat) :: x
        INTEGER*8 :: prec
        prec = c_prec
        CALL exflib_float_ltoef_FRT(x%num, a, prec)
      END FUNCTION exflib_float_qqtoef_p

      ! int to exfloat (function form)
      FUNCTION exflib_float_itoef_p(a) RESULT(x)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat) :: x
        INTEGER*8 :: prec, aa
        aa = a
        prec = c_prec
        CALL exflib_float_ltoef_FRT(x%num, aa, prec)
      END FUNCTION exflib_float_itoef_p

      ! double to exfloat
      FUNCTION exflib_float_dtoef_p(d) RESULT(x)
        REAL*8, INTENT(IN) :: d
        TYPE(exfloat) :: x
        INTEGER*8 :: prec
        prec = c_prec
        CALL exflib_float_dtoef_FRT(x%num, d, prec)
      END FUNCTION exflib_float_dtoef_p

      ! single to exfloat
      FUNCTION exflib_float_stoef_p(s) RESULT(x)
        REAL*4, INTENT(IN) :: s
        TYPE(exfloat) :: x
        INTEGER*8 :: prec
        REAL*8 :: d
        d = s
        prec = c_prec
        CALL exflib_float_dtoef_FRT(x%num, d, prec)
      END FUNCTION exflib_float_stoef_p

      ! expression to ezfloat
      SUBROUTINE exflib_float_etoez(z, a)
        TYPE(ezfloat), INTENT(OUT) :: z
        CHARACTER(*), INTENT(IN) :: a
        z%re = a
        z%im = 0
      END SUBROUTINE exflib_float_etoez

      ! exfloat to ezfloat
      SUBROUTINE exflib_float_eftoez(z, a)
        TYPE(ezfloat), INTENT(OUT) :: z
        TYPE(exfloat), INTENT(IN) :: a
        z%re = a
        z%im = 0
      END SUBROUTINE exflib_float_eftoez

      ! quadword to ezfloat
      SUBROUTINE exflib_float_qtoez(z, a)
        TYPE(ezfloat), INTENT(OUT) :: z
        INTEGER*8, INTENT(IN) :: a
        z%re = a
        z%im = 0
      END SUBROUTINE exflib_float_qtoez

      ! int to ezfloat
      SUBROUTINE exflib_float_itoez(z, a)
        TYPE(ezfloat), INTENT(OUT) :: z
        INTEGER*4, INTENT(IN) :: a
        z%re = a
        z%im = 0
      END SUBROUTINE exflib_float_itoez

      ! ezfloat to single complex
      SUBROUTINE exflib_float_eztoc(z, a)
        COMPLEX*8, INTENT(OUT) :: z
        TYPE(ezfloat), INTENT(IN) :: a
        REAL :: re, im
        re = a%re
        im = a%im
        z = CMPLX(re, im)
      END SUBROUTINE exflib_float_eztoc

      ! ezfloat to double complex
      SUBROUTINE exflib_float_eztow(z, a)
        COMPLEX*16,INTENT(OUT) :: z
        TYPE(ezfloat), INTENT(IN) :: a
        REAL*8 :: re, im
        re = a%re
        im = a%im
        z = DCMPLX(re, im)
      END SUBROUTINE exflib_float_eztow

      ! dcmplx to ezfloat
      FUNCTION exflib_float_wtoez_p(d) RESULT(z)
        COMPLEX*16, INTENT(IN) :: d
        TYPE(ezfloat) :: z
        z%re = exflib_float_dtoef_p( REAL(d) )
        z%im = exflib_float_dtoef_p( AIMAG(d) )
      END FUNCTION exflib_float_wtoez_p

      ! smcplx to ezfloat
      FUNCTION exflib_float_ctoez_p(s) RESULT(z)
        COMPLEX*8, INTENT(IN) :: s
        TYPE(ezfloat) :: z
        z%re = exflib_float_stoef_p( REAL(s) )
        z%im = exflib_float_stoef_p( AIMAG(s) )
      END FUNCTION exflib_float_ctoez_p

!      ! double to ezfloat
!      FUNCTION exflib_float_dtoez_p(d) RESULT(z)
!        REAL*8, INTENT(IN) :: d
!        TYPE(ezfloat) :: z
!        z%re = exflib_float_dtoef_p( d )
!        z%im = 0
!      END FUNCTION exflib_float_dtoez_p
!
!      ! single to ezfloat
!      FUNCTION exflib_float_stoez_p(s) RESULT(z)
!        REAL*4, INTENT(IN) :: s
!        TYPE(ezfloat) :: z
!        z%re = exflib_float_stoef_p( s )
!        z%im = 0
!      END FUNCTION exflib_float_stoez_p

      !------------------------------------------------------------
      ! cnversion to formatted string
      !------------------------------------------------------------
      FUNCTION exflib_format_f(fmt, a) RESULT(str)
        CHARACTER(*), INTENT(IN) :: fmt         ! CHARACTER(:) is error
        TYPE(exfloat), INTENT(IN) :: a
        CHARACTER(max_string_length) :: str     ! ALLOCATABLE is desired
                                                ! indicating index from 0 ?
        INTEGER*8 :: prec, dst_len
        prec = c_prec
        dst_len = len(str)
        str = ''
        CALL exflib_float_fortran_format_FRT(str, fmt//ACHAR(0), a%num, prec, dst_len)
      END FUNCTION exflib_format_f

      FUNCTION exflib_format_zf(fmt, a) RESULT(str)
        CHARACTER(*), INTENT(IN) :: fmt
        TYPE(ezfloat), INTENT(IN) :: a
        CHARACTER(max_string_length) :: str
        str = ''
        str = '(' // TRIM(exflib_format_f(fmt, a%re)) // ',' &
             // TRIM(exflib_format_f(fmt, a%im)) // ')'
      END FUNCTION exflib_format_zf

      !SUBROUTINE exflib_format_p(str, fmt, a)
      !CHARACTER(max_string_length), INTENT(OUT) :: str
      !  CHARACTER(*), INTENT(IN) :: fmt
      !  TYPE(exfloat), INTENT(IN) :: a
      !
      !  INTEGER*8 :: prec, dst_len
      !  prec = c_prec
      !  dst_len = len(str)
      !  str = ''
      !  CALL exflib_float_fortran_format_FRT(str, fmt//ACHAR(0), a%num, prec, dst_len)
      !END SUBROUTINE exflib_format_p

      !------------------------------------------------------------
      ! relation, == (equal)
      !------------------------------------------------------------
      ! exfloat == exfloat
      FUNCTION exflib_float_eq_mm(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a, b
        LOGICAL*8 :: x
        INTEGER*8 :: i
        x = .TRUE.

        ! zero detection
        IF ( (a%num(0)*2 == 0) .AND. (b%num(0)*2 == 0) ) THEN
           RETURN
        END IF

        DO i = 0, c_prec
           IF ( a%num(i) /= b%num(i) ) THEN
              x = .FALSE.
           ENDIF
        ENDDO
      END FUNCTION exflib_float_eq_mm

      ! exfloat == quadword
      FUNCTION exflib_float_eq_mq(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8 :: x
        TYPE(exfloat) :: tmp
        INTEGER*8 :: prec
        prec = c_prec
        CALL exflib_float_ltoef_FRT(tmp%num, b, prec)
        x = exflib_float_eq_mm(a, tmp)
      END FUNCTION exflib_float_eq_mq

      ! exfloat == int
      FUNCTION exflib_float_eq_mi(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8 :: x
        TYPE(exfloat) :: tmp
        INTEGER*8 :: prec, bb
        bb = b
        prec = c_prec
        CALL exflib_float_ltoef_FRT(tmp%num, bb, prec)
        x = exflib_float_eq_mm(a, tmp)
      END FUNCTION exflib_float_eq_mi

      ! exfloat == double
      FUNCTION exflib_float_eq_md(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8 :: x
        TYPE(exfloat) :: tmp
        INTEGER*8 :: prec
        prec = c_prec
        CALL exflib_float_dtoef_FRT(tmp%num, b, prec)
        x = exflib_float_eq_mm(a, tmp)
      END FUNCTION exflib_float_eq_md

      ! exfloat == single
      FUNCTION exflib_float_eq_ms(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8 :: x
        TYPE(exfloat) :: tmp
        INTEGER*8 :: prec
        REAL*8 :: bb
        bb = b
        prec = c_prec
        CALL exflib_float_dtoef_FRT(tmp%num, bb, prec)
        x = exflib_float_eq_mm(a, tmp)
      END FUNCTION exflib_float_eq_ms

      ! quadword == exfloat
      FUNCTION exflib_float_eq_qm(a, b) RESULT(x)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        TYPE(exfloat) :: tmp
        INTEGER*8 :: prec
        prec = c_prec
        CALL exflib_float_ltoef_FRT(tmp%num, a, prec)
        x = exflib_float_eq_mm(tmp, b)
      END FUNCTION exflib_float_eq_qm

      ! int == exfloat
      FUNCTION exflib_float_eq_im(a, b) RESULT(x)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        TYPE(exfloat) :: tmp
        INTEGER*8 :: prec, aa
        aa = a
        prec = c_prec
        CALL exflib_float_ltoef_FRT(tmp%num, aa, prec)
        x = exflib_float_eq_mm(tmp, b)
      END FUNCTION exflib_float_eq_im

      ! double == exfloat
      FUNCTION exflib_float_eq_dm(a, b) RESULT(x)
        REAL*8, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        TYPE(exfloat) :: tmp
        INTEGER*8 :: prec
        prec = c_prec
        CALL exflib_float_dtoef_FRT(tmp%num, a, prec)
        x = exflib_float_eq_mm(tmp, b)
      END FUNCTION exflib_float_eq_dm

      ! single == exfloat
      FUNCTION exflib_float_eq_sm(a, b) RESULT(x)
        REAL*4, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        TYPE(exfloat) :: tmp
        INTEGER*8 :: prec
        REAL*8 :: aa
        aa = a
        prec = c_prec
        CALL exflib_float_dtoef_FRT(tmp%num, aa, prec)
        x = exflib_float_eq_mm(tmp, b)
      END FUNCTION exflib_float_eq_sm

      ! exfloat == dcomplex
      FUNCTION exflib_float_eq_xw(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        COMPLEX*16, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .FALSE.
        IF ( (a == REAL(b)).AND.(0 == AIMAG(b)) ) THEN
           x = .TRUE.
        END IF
      END FUNCTION exflib_float_eq_xw

      ! dcomplex == exfloat
      FUNCTION exflib_float_eq_wx(a, b) RESULT(x)
        COMPLEX*16, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .FALSE.
        IF ( (REAL(a) == b).AND.(AIMAG(a) == 0) ) THEN
           x = .TRUE.
        END IF
      END FUNCTION exflib_float_eq_wx

      ! exfloat == scomplex
      FUNCTION exflib_float_eq_xc(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        COMPLEX*8, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .FALSE.
        IF ( (a == REAL(b)).AND.(0 == AIMAG(b)) ) THEN
           x = .TRUE.
        END IF
      END FUNCTION exflib_float_eq_xc

      ! scomplex == exfloat
      FUNCTION exflib_float_eq_cx(a, b) RESULT(x)
        COMPLEX*8, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .FALSE.
        IF ( (REAL(a) == b).AND.(AIMAG(a) == 0) ) THEN
           x = .TRUE.
        END IF
      END FUNCTION exflib_float_eq_cx

      ! ezfloat == ezfloat
      FUNCTION exflib_float_eq_zz(a, b) RESULT(x)
        TYPE(ezfloat), INTENT(IN) :: a, b
        LOGICAL*8 :: x
        x = .FALSE.
        IF ( (a%re == b%re).AND.(a%im == b%im) ) THEN
           x = .TRUE.
        END IF
      END FUNCTION exflib_float_eq_zz

      ! ezfloat == dcomplex
      FUNCTION exflib_float_eq_zw(a, b) RESULT(x)
        TYPE(ezfloat), INTENT(IN) :: a
        COMPLEX*16, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .FALSE.
        IF ( (a%re == REAL(b)).AND.(a%im == AIMAG(b)) ) THEN
           x = .TRUE.
        END IF
      END FUNCTION exflib_float_eq_zw

      ! dcomplex == ezfloat
      FUNCTION exflib_float_eq_wz(a, b) RESULT(x)
        COMPLEX*16, INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .FALSE.
        IF ( (REAL(a) == b%re).AND.(AIMAG(a) == b%im) ) THEN
           x = .TRUE.
        END IF
      END FUNCTION exflib_float_eq_wz

      ! ezfloat == scomplex
      FUNCTION exflib_float_eq_zc(a, b) RESULT(x)
        TYPE(ezfloat), INTENT(IN) :: a
        COMPLEX*8, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .FALSE.
        IF ( (a%re == REAL(b)).AND.(a%im == AIMAG(b)) ) THEN
           x = .TRUE.
        END IF
      END FUNCTION exflib_float_eq_zc

      ! scomplex == ezfloat
      FUNCTION exflib_float_eq_cz(a, b) RESULT(x)
        COMPLEX*8, INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .FALSE.
        IF ( (REAL(a) == b%re).AND.(AIMAG(a) == b%im) ) THEN
           x = .TRUE.
        END IF
      END FUNCTION exflib_float_eq_cz

      ! ezfloat == exfloat
      FUNCTION exflib_float_eq_zm(a, b) RESULT(x)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .FALSE.
        IF ( (a%re == b).AND.(a%im == 0) ) THEN
           x = .TRUE.
        END IF
      END FUNCTION exflib_float_eq_zm

      ! exfloat == ezfloat
      FUNCTION exflib_float_eq_mz(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .FALSE.
        IF ( (b%re == a).AND.(b%im == 0) ) THEN
           x = .TRUE.
        END IF
      END FUNCTION exflib_float_eq_mz

      ! ezfloat == quadword
      FUNCTION exflib_float_eq_zq(a, b) RESULT(x)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .FALSE.
        IF ( (a%re == b).AND.(a%im == 0) ) THEN
           x = .TRUE.
        END IF
      END FUNCTION exflib_float_eq_zq

      ! ezfloat == int
      FUNCTION exflib_float_eq_zi(a, b) RESULT(x)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .FALSE.
        IF ( (a%re == b).AND.(a%im == 0) ) THEN
           x = .TRUE.
        END IF
      END FUNCTION exflib_float_eq_zi

      ! ezfloat == double
      FUNCTION exflib_float_eq_zd(a, b) RESULT(x)
        TYPE(ezfloat), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .FALSE.
        IF ( (a%re == b).AND.(a%im == 0) ) THEN
           x = .TRUE.
        END IF
      END FUNCTION exflib_float_eq_zd

      ! ezfloat == single
      FUNCTION exflib_float_eq_zs(a, b) RESULT(x)
        TYPE(ezfloat), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .FALSE.
        IF ( (a%re == b).AND.(a%im == 0) ) THEN
           x = .TRUE.
        END IF
      END FUNCTION exflib_float_eq_zs

      ! quadword == ezfloat
      FUNCTION exflib_float_eq_qz(a, b) RESULT(x)
        INTEGER*8, INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .FALSE.
        IF ( (b%re == a).AND.(b%im == 0) ) THEN
           x = .TRUE.
        END IF
      END FUNCTION exflib_float_eq_qz

      ! int == ezfloat
      FUNCTION exflib_float_eq_iz(a, b) RESULT(x)
        INTEGER*4, INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .FALSE.
        IF ( (b%re == a).AND.(b%im == 0) ) THEN
           x = .TRUE.
        END IF
      END FUNCTION exflib_float_eq_iz

      ! double == ezfloat
      FUNCTION exflib_float_eq_w(a, b) RESULT(x)
        REAL*8, INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .FALSE.
        IF ( (b%re == a).AND.(b%im == 0) ) THEN
           x = .TRUE.
        END IF
      END FUNCTION exflib_float_eq_w

      ! single == ezfloat
      FUNCTION exflib_float_eq_c(a, b) RESULT(x)
        REAL*4, INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .FALSE.
        IF ( (b%re == a).AND.(b%im == 0) ) THEN
           x = .TRUE.
        END IF
      END FUNCTION exflib_float_eq_c

      !------------------------------------------------------------
      ! relation, /= (not equal)
      !------------------------------------------------------------
      ! exfloat /= exfloat
      FUNCTION exflib_float_ne_mm(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a, b
        LOGICAL*8 :: x
        x = .NOT. (exflib_float_eq_mm(a, b))
      END FUNCTION exflib_float_ne_mm

      ! exfloat /= quadword
      FUNCTION exflib_float_ne_mq(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (exflib_float_eq_mq(a, b))
      END FUNCTION exflib_float_ne_mq

      ! exfloat /= int
      FUNCTION exflib_float_ne_mi(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (exflib_float_eq_mi(a, b))
      END FUNCTION exflib_float_ne_mi

      ! exfloat /= double
      FUNCTION exflib_float_ne_md(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (exflib_float_eq_md(a, b))
      END FUNCTION exflib_float_ne_md

      ! exfloat /= single
      FUNCTION exflib_float_ne_ms(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (exflib_float_eq_ms(a, b))
      END FUNCTION exflib_float_ne_ms

      ! quadword /= exfloat
      FUNCTION exflib_float_ne_qm(a, b) RESULT(x)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (exflib_float_eq_qm(a, b))
      END FUNCTION exflib_float_ne_qm

      ! int /= exfloat
      FUNCTION exflib_float_ne_im(a, b) RESULT(x)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (exflib_float_eq_im(a, b))
      END FUNCTION exflib_float_ne_im

      ! double /= exfloat
      FUNCTION exflib_float_ne_dm(a, b) RESULT(x)
        REAL*8, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (exflib_float_eq_dm(a, b))
      END FUNCTION exflib_float_ne_dm

      ! single /= exfloat
      FUNCTION exflib_float_ne_sm(a, b) RESULT(x)
        REAL*4, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (exflib_float_eq_sm(a, b))
      END FUNCTION exflib_float_ne_sm

      ! exfloat /= dcomplex
      FUNCTION exflib_float_ne_xw(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        COMPLEX*16, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (a == b)
      END FUNCTION exflib_float_ne_xw

      ! dcomplex /= exfloat
      FUNCTION exflib_float_ne_wx(a, b) RESULT(x)
        COMPLEX*16, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (a == b)
      END FUNCTION exflib_float_ne_wx

      ! exfloat /= scomplex
      FUNCTION exflib_float_ne_xc(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        COMPLEX*8, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (a == b)
      END FUNCTION exflib_float_ne_xc

      ! scomplex /= exfloat
      FUNCTION exflib_float_ne_cx(a, b) RESULT(x)
        COMPLEX*8, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (a == b)
      END FUNCTION exflib_float_ne_cx

      ! ezfloat /= ezfloat
      FUNCTION exflib_float_ne_zz(a, b) RESULT(x)
        TYPE(ezfloat), INTENT(IN) :: a, b
        LOGICAL*8 :: x
        x = .NOT. (a == b)
      END FUNCTION exflib_float_ne_zz

      ! ezfloat /= dcomplex
      FUNCTION exflib_float_ne_zw(a, b) RESULT(x)
        TYPE(ezfloat), INTENT(IN) :: a
        COMPLEX*16, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (a == b)
      END FUNCTION exflib_float_ne_zw

      ! dcomplex /= ezfloat
      FUNCTION exflib_float_ne_wz(a, b) RESULT(x)
        COMPLEX*16, INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (a == b)
      END FUNCTION exflib_float_ne_wz

      ! ezfloat /= scomplex
      FUNCTION exflib_float_ne_zc(a, b) RESULT(x)
        TYPE(ezfloat), INTENT(IN) :: a
        COMPLEX*8, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (a == b)
      END FUNCTION exflib_float_ne_zc

      ! scomplex /= ezfloat
      FUNCTION exflib_float_ne_cz(a, b) RESULT(x)
        COMPLEX*8, INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (a == b)
      END FUNCTION exflib_float_ne_cz

      ! ezfloat /= exfloat
      FUNCTION exflib_float_ne_zm(a, b) RESULT(x)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (a == b)
      END FUNCTION exflib_float_ne_zm

      ! exfloat /= ezfloat
      FUNCTION exflib_float_ne_mz(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (a == b)
      END FUNCTION exflib_float_ne_mz

      ! ezfloat /= quadword
      FUNCTION exflib_float_ne_zq(a, b) RESULT(x)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (a == b)
      END FUNCTION exflib_float_ne_zq

      ! ezfloat /= int
      FUNCTION exflib_float_ne_zi(a, b) RESULT(x)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (a == b)
      END FUNCTION exflib_float_ne_zi

      ! ezfloat /= double
      FUNCTION exflib_float_ne_zd(a, b) RESULT(x)
        TYPE(ezfloat), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (a == b)
      END FUNCTION exflib_float_ne_zd

      ! ezfloat /= single
      FUNCTION exflib_float_ne_zs(a, b) RESULT(x)
        TYPE(ezfloat), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (a == b)
      END FUNCTION exflib_float_ne_zs

      ! quadword /= ezfloat
      FUNCTION exflib_float_ne_qz(a, b) RESULT(x)
        INTEGER*8, INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (a == b)
      END FUNCTION exflib_float_ne_qz

      ! int /= ezfloat
      FUNCTION exflib_float_ne_iz(a, b) RESULT(x)
        INTEGER*4, INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (a == b)
      END FUNCTION exflib_float_ne_iz

      ! double /= ezfloat
      FUNCTION exflib_float_ne_dz(a, b) RESULT(x)
        REAL*8, INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (a == b)
      END FUNCTION exflib_float_ne_dz

      ! single /= ezfloat
      FUNCTION exflib_float_ne_sz(a, b) RESULT(x)
        REAL*4, INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT. (a == b)
      END FUNCTION exflib_float_ne_sz

      !------------------------------------------------------------
      ! relation, < (less than)
      !------------------------------------------------------------
      ! exfloat < exfloat
      FUNCTION exflib_float_lt_mm(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a, b
        LOGICAL*8 :: x
        INTEGER*8 :: prec
        prec = c_prec
        x = exflib_float_lt_FRT(a%num, b%num, prec)
      END FUNCTION exflib_float_lt_mm

      ! exfloat < quadword
      FUNCTION exflib_float_lt_mq(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8 :: x
        INTEGER*8 :: prec
        TYPE(exfloat) :: tmp
        prec = c_prec
        CALL exflib_float_ltoef_FRT(tmp%num, b, prec)
        prec = c_prec
        x = exflib_float_lt_FRT(a%num, tmp%num, prec)
      END FUNCTION exflib_float_lt_mq

      ! exfloat < int
      FUNCTION exflib_float_lt_mi(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8 :: x
        INTEGER*8 :: prec, bb
        TYPE(exfloat) :: tmp
        bb = b
        prec = c_prec
        CALL exflib_float_ltoef_FRT(tmp%num, bb, prec)
        prec = c_prec
        x = exflib_float_lt_FRT(a%num, tmp%num, prec)
      END FUNCTION exflib_float_lt_mi

      ! exfloat < double
      FUNCTION exflib_float_lt_md(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8 :: x
        INTEGER*8 :: prec
        TYPE(exfloat) :: tmp
        prec = c_prec
        CALL exflib_float_dtoef_FRT(tmp%num, b, prec)
        prec = c_prec
        x = exflib_float_lt_FRT(a%num, tmp%num, prec)
      END FUNCTION exflib_float_lt_md

      ! exfloat < single
      FUNCTION exflib_float_lt_ms(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8 :: x
        INTEGER*8 :: prec
        TYPE(exfloat) :: tmp
        REAL*8 bb
        bb = b
        prec = c_prec
        CALL exflib_float_dtoef_FRT(tmp%num, bb, prec)
        prec = c_prec
        x = exflib_float_lt_FRT(a%num, tmp%num, prec)
      END FUNCTION exflib_float_lt_ms

      ! quadword < exfloat
      FUNCTION exflib_float_lt_qm(a, b) RESULT(x)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        TYPE(exfloat) :: tmp
        INTEGER*8 :: prec
        prec = c_prec
        CALL exflib_float_ltoef_FRT(tmp%num, a, prec)
        prec = c_prec
        x = exflib_float_lt_FRT(tmp%num, b%num, prec)
      END FUNCTION exflib_float_lt_qm

      ! int < exfloat
      FUNCTION exflib_float_lt_im(a, b) RESULT(x)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        TYPE(exfloat) :: tmp
        INTEGER*8 :: prec, aa
        aa = a
        prec = c_prec
        CALL exflib_float_ltoef_FRT(tmp%num, aa, prec)
        prec = c_prec
        x = exflib_float_lt_FRT(tmp%num, b%num, prec)
      END FUNCTION exflib_float_lt_im

      ! double < exfloat
      FUNCTION exflib_float_lt_dm(a, b) RESULT(x)
        REAL*8, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        INTEGER*8 :: prec
        TYPE(exfloat) :: tmp
        prec = c_prec
        CALL exflib_float_dtoef_FRT(tmp%num, a, prec)
        prec = c_prec
        x = exflib_float_lt_FRT(tmp%num, b%num, prec)
      END FUNCTION exflib_float_lt_dm

      ! single < exfloat
      FUNCTION exflib_float_lt_sm(a, b) RESULT(x)
        REAL*4, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        INTEGER*8 :: prec
        TYPE(exfloat) :: tmp
        REAL*8 :: aa
        aa = a
        prec = c_prec
        CALL exflib_float_dtoef_FRT(tmp%num, aa, prec)
        prec = c_prec
        x = exflib_float_lt_FRT(tmp%num, b%num, prec)
      END FUNCTION exflib_float_lt_sm

      !------------------------------------------------------------
      ! relation, > (greater than)
      !------------------------------------------------------------
      ! exfloat > exfloat
      FUNCTION exflib_float_gt_mm(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a, b
        LOGICAL*8 :: x
        INTEGER*8 :: prec
        prec = c_prec
        x = exflib_float_lt_FRT(b%num, a%num, prec)
      END FUNCTION exflib_float_gt_mm

      ! exfloat > quadword
      FUNCTION exflib_float_gt_mq(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8 :: x
        INTEGER*8 :: prec
        TYPE(exfloat) :: tmp
        prec = c_prec
        CALL exflib_float_ltoef_FRT(tmp%num, b, prec)
        prec = c_prec
        x = exflib_float_lt_FRT(tmp%num, a%num, prec)
      END FUNCTION exflib_float_gt_mq

      ! exfloat > int
      FUNCTION exflib_float_gt_mi(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8 :: x
        INTEGER*8 :: prec, bb
        TYPE(exfloat) :: tmp
        bb = b
        prec = c_prec
        CALL exflib_float_ltoef_FRT(tmp%num, bb, prec)
        prec = c_prec
        x = exflib_float_lt_FRT(tmp%num, a%num, prec)
      END FUNCTION exflib_float_gt_mi

      ! exfloat > double
      FUNCTION exflib_float_gt_md(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8 :: x
        INTEGER*8 :: prec
        TYPE(exfloat) :: tmp
        prec = c_prec
        CALL exflib_float_dtoef_FRT(tmp%num, b, prec)
        prec = c_prec
        x = exflib_float_lt_FRT(tmp%num, a%num, prec)
      END FUNCTION exflib_float_gt_md

      ! exfloat > single
      FUNCTION exflib_float_gt_ms(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8 :: x
        INTEGER*8 :: prec
        TYPE(exfloat) :: tmp
        REAL*8 bb
        bb = b
        prec = c_prec
        CALL exflib_float_dtoef_FRT(tmp%num, bb, prec)
        prec = c_prec
        x = exflib_float_lt_FRT(tmp%num, a%num, prec)
      END FUNCTION exflib_float_gt_ms

      ! quadword > exfloat
      FUNCTION exflib_float_gt_qm(a, b) RESULT(x)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        TYPE(exfloat) :: tmp
        INTEGER*8 :: prec
        prec = c_prec
        CALL exflib_float_ltoef_FRT(tmp%num, a, prec)
        prec = c_prec
        x = exflib_float_lt_FRT(b%num, tmp%num, prec)
      END FUNCTION exflib_float_gt_qm

      ! int > exfloat
      FUNCTION exflib_float_gt_im(a, b) RESULT(x)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        TYPE(exfloat) :: tmp
        INTEGER*8 :: prec, aa
        aa = a
        prec = c_prec
        CALL exflib_float_ltoef_FRT(tmp%num, aa, prec)
        prec = c_prec
        x = exflib_float_lt_FRT(b%num, tmp%num, prec)
      END FUNCTION exflib_float_gt_im

      ! double > exfloat
      FUNCTION exflib_float_gt_dm(a, b) RESULT(x)
        REAL*8, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        INTEGER*8 :: prec
        TYPE(exfloat) :: tmp
        prec = c_prec
        CALL exflib_float_dtoef_FRT(tmp%num, a, prec)
        prec = c_prec
        x = exflib_float_lt_FRT(b%num, tmp%num, prec)
      END FUNCTION exflib_float_gt_dm

      ! single > exfloat
      FUNCTION exflib_float_gt_sm(a, b) RESULT(x)
        REAL*4, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        INTEGER*8 :: prec
        TYPE(exfloat) :: tmp
        REAL*8 :: aa
        aa = a
        prec = c_prec
        CALL exflib_float_dtoef_FRT(tmp%num, aa, prec)
        prec = c_prec
        x = exflib_float_lt_FRT(b%num, tmp%num, prec)
      END FUNCTION exflib_float_gt_sm

      !------------------------------------------------------------
      ! relation, >= (greater than or equal)
      !------------------------------------------------------------
      ! exfloat >= exfloat
      FUNCTION exflib_float_ge_mm(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a, b
        LOGICAL*8 :: x
        x = .NOT.(exflib_float_lt_mm(a, b))
      END FUNCTION exflib_float_ge_mm

      ! exfloat >= quadword
      FUNCTION exflib_float_ge_mq(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT.(exflib_float_lt_mq(a, b))
      END FUNCTION exflib_float_ge_mq

      ! exfloat >= int
      FUNCTION exflib_float_ge_mi(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT.(exflib_float_lt_mi(a, b))
      END FUNCTION exflib_float_ge_mi

      ! exfloat >= double
      FUNCTION exflib_float_ge_md(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT.(exflib_float_lt_md(a, b))
      END FUNCTION exflib_float_ge_md

      ! exfloat >= single
      FUNCTION exflib_float_ge_ms(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT.(exflib_float_lt_ms(a, b))
      END FUNCTION exflib_float_ge_ms

      ! quadword >= exfloat
      FUNCTION exflib_float_ge_qm(a, b) RESULT(x)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT.(exflib_float_lt_qm(a, b))
      END FUNCTION exflib_float_ge_qm

      ! int >= exfloat
      FUNCTION exflib_float_ge_im(a, b) RESULT(x)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT.(exflib_float_lt_im(a, b))
      END FUNCTION exflib_float_ge_im

      ! double >= exfloat
      FUNCTION exflib_float_ge_dm(a, b) RESULT(x)
        REAL*8, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT.(exflib_float_lt_dm(a, b))
      END FUNCTION exflib_float_ge_dm

      ! single >= exfloat
      FUNCTION exflib_float_ge_sm(a, b) RESULT(x)
        REAL*4, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT.(exflib_float_lt_sm(a, b))
      END FUNCTION exflib_float_ge_sm

      !------------------------------------------------------------
      ! relation, <= (less than or equal)
      !------------------------------------------------------------
      ! exfloat <= exfloat
      FUNCTION exflib_float_le_mm(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a, b
        LOGICAL*8 :: x
        x = .NOT.(exflib_float_gt_mm(a, b))
      END FUNCTION exflib_float_le_mm

      ! exfloat <= quadword
      FUNCTION exflib_float_le_mq(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT.(exflib_float_gt_mq(a, b))
      END FUNCTION exflib_float_le_mq

      ! exfloat <= int
      FUNCTION exflib_float_le_mi(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT.(exflib_float_gt_mi(a, b))
      END FUNCTION exflib_float_le_mi

      ! exfloat <= double
      FUNCTION exflib_float_le_md(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*8, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT.(exflib_float_gt_md(a, b))
      END FUNCTION exflib_float_le_md

      ! exfloat <= single
      FUNCTION exflib_float_le_ms(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        REAL*4, INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT.(exflib_float_gt_ms(a, b))
      END FUNCTION exflib_float_le_ms

      ! quadword <= exfloat
      FUNCTION exflib_float_le_qm(a, b) RESULT(x)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT.(exflib_float_gt_qm(a, b))
      END FUNCTION exflib_float_le_qm

      ! int <= exfloat
      FUNCTION exflib_float_le_im(a, b) RESULT(x)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT.(exflib_float_gt_im(a, b))
      END FUNCTION exflib_float_le_im

      ! double <= exfloat
      FUNCTION exflib_float_le_dm(a, b) RESULT(x)
        REAL*8, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT.(exflib_float_gt_dm(a, b))
      END FUNCTION exflib_float_le_dm

      ! single <= exfloat
      FUNCTION exflib_float_le_sm(a, b) RESULT(x)
        REAL*4, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        LOGICAL*8 :: x
        x = .NOT.(exflib_float_gt_sm(a, b))
      END FUNCTION exflib_float_le_sm

      !------------------------------------------------------------
      ! mathematical function
      !------------------------------------------------------------
      FUNCTION exflib_float_abs_f(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
        INTEGER*8 :: prec
        prec = c_prec
        CALL exflib_float_fabs_FRT(x%num, a%num, prec)
      END FUNCTION exflib_float_abs_f

      FUNCTION exflib_float_sqrt_f(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
        INTEGER*8 :: prec, hint
        INTEGER*8 :: workmem(0:work_size)
        prec = c_prec
        hint = exflib_float_sqrt_FRT(x%num, a%num, prec, workmem)
        prec = c_prec
	CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_sqrt_f

!      FUNCTION exflib_float_cubic_f(a) RESULT(x)
!        TYPE(exfloat), INTENT(IN) :: a
!        TYPE(exfloat) :: x
!
!        TODO
!
!      END FUNCTION exflib_float_cubic_f

      FUNCTION exflib_float_sin_f(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
        INTEGER*8 :: prec, hint
        INTEGER*8 :: workmem(0:work_size)
        prec = c_prec
        hint = exflib_float_sin_FRT(x%num, a%num, prec, workmem)
        prec = c_prec
	CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_sin_f

      FUNCTION exflib_float_cos_f(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
        INTEGER*8 :: prec, hint
        INTEGER*8 :: workmem(0:work_size)
        prec = c_prec
        hint = exflib_float_cos_FRT(x%num, a%num, prec, workmem)
        prec = c_prec
	CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_cos_f

      FUNCTION exflib_float_tan_f(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
        x = sin(a)/cos(a)
      END FUNCTION exflib_float_tan_f

      FUNCTION exflib_float_exp_f(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
        INTEGER*8 :: prec, hint
        INTEGER*8 :: workmem(0:work_size)
        prec = c_prec
        hint = exflib_float_exp_FRT(x%num, a%num, prec, workmem)
        prec = c_prec
	CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_exp_f

      ! exfloat ** exfloat
      FUNCTION exflib_float_pow_mm(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a, b
        TYPE(exfloat) :: x
        INTEGER*8 :: prec, hint, i
        INTEGER*8 :: workmem(0:work_size)

        !! comment out on 2017.09.23 (detecting in func_pow.c)
        !! inserted on 2016.09.30
        !! zero and non-zero detection
        !IF ( (a%num(0)*2 == 0) .AND. (b%num(0)*2 /= 0) ) THEN
        !   DO i = 0, c_prec
        !      x%num(i) = 0
        !   END DO
        !   RETURN
        !END IF

        prec = c_prec
#ifdef _OPENMP
        !$OMP CRITICAL ( EXFLIB_INV_TABLE_LOG )
        IF ( inv_table_log(0) == 0 ) THEN

           CALL exflib_init_reciprocal_table_FRT( inv_table_log, prec+2 )

        END IF
        !$OMP END CRITICAL ( EXFLIB_INV_TABLE_LOG )
#endif
        hint = exflib_float_pow_FRT(x%num, a%num, b%num, prec, workmem, inv_table_log)
        prec = c_prec
        CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_pow_mm

      ! exfloat ** quadword
      FUNCTION exflib_float_pow_mq(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(exfloat) :: x
        INTEGER*8 :: prec, hint
        INTEGER*8 :: workmem(0:work_size)
        prec = c_prec
        hint = exflib_float_pow_long_FRT(x%num, a%num, b, prec, workmem)
        prec = c_prec
        CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_pow_mq

      ! exfloat ** int
      FUNCTION exflib_float_pow_mi(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(exfloat) :: x
        INTEGER*8 :: prec, hint, bb
        INTEGER*8 :: workmem(0:work_size)
        prec = c_prec
        bb = b
        hint = exflib_float_pow_long_FRT(x%num, a%num, bb, prec, workmem)
        prec = c_prec
        CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_pow_mi

      ! quadword ** exfloat
      FUNCTION exflib_float_pow_qm(a, b) RESULT(x)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat) :: x
        TYPE(exfloat) :: aa
        INTEGER*8 :: prec, hint
        INTEGER*8 :: workmem(0:work_size)
        prec = c_prec
        CALL exflib_float_ltoef_FRT(aa%num, a, prec)
        prec = c_prec
#ifdef _OPENMP
        !$OMP CRITICAL ( EXFLIB_INV_TABLE_LOG )
        IF ( inv_table_log(0) == 0 ) THEN

           CALL exflib_init_reciprocal_table_FRT( inv_table_log, prec+2 )

        END IF
        !$OMP END CRITICAL ( EXFLIB_INV_TABLE_LOG )
#endif
        hint = exflib_float_pow_FRT(x%num, aa%num, b%num, prec, workmem, inv_table_log)
        prec = c_prec
        CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_pow_qm

      ! int ** exfloat
      FUNCTION exflib_float_pow_im(a, b) RESULT(x)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(exfloat) :: x
        TYPE(exfloat) :: aa
        INTEGER*8 :: prec, hint, a_long
        INTEGER*8 :: workmem(0:work_size)
        prec = c_prec
        a_long = a
        CALL exflib_float_ltoef_FRT(aa%num, a_long, prec)
        prec = c_prec
#ifdef _OPENMP
        !$OMP CRITICAL ( EXFLIB_INV_TABLE_LOG )
        IF ( inv_table_log(0) == 0 ) THEN

           CALL exflib_init_reciprocal_table_FRT( inv_table_log, prec+2 )

        END IF
        !$OMP END CRITICAL ( EXFLIB_INV_TABLE_LOG )
#endif
        hint = exflib_float_pow_FRT(x%num, aa%num, b%num, prec, workmem, inv_table_log)
        prec = c_prec
        CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_pow_im

      FUNCTION exflib_float_log_f(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
        INTEGER*8 :: prec, hint
        INTEGER*8 :: workmem(0:work_size)
        prec = c_prec
#ifdef _OPENMP
        !$OMP CRITICAL ( EXFLIB_INV_TABLE_LOG )
        IF ( inv_table_log(0) == 0 ) THEN

           CALL exflib_init_reciprocal_table_FRT( inv_table_log, prec+2 )

        END IF
        !$OMP END CRITICAL ( EXFLIB_INV_TABLE_LOG )
#endif
        hint = exflib_float_log_FRT(x%num, a%num, prec, workmem, inv_table_log)
        prec = c_prec
	CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_log_f

      FUNCTION exflib_float_log10_f(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
	TYPE(exfloat) :: ln10
	INTEGER*8 :: prec
	prec = c_prec
	CALL exflib_float_ctoef_FRT(ln10%num, 'ln10' // ACHAR(0), prec)
	x = log(a)
	x = x / ln10
      END FUNCTION exflib_float_log10_f

      FUNCTION exflib_float_sinh_f(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
	TYPE(exfloat) :: e1, e2
	e1 = exp(a)
	e2 = 1/e1
	x = (e1 - e2) / 2
      END FUNCTION exflib_float_sinh_f

      FUNCTION exflib_float_cosh_f(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
	TYPE(exfloat) :: e1, e2
	e1 = exp(a)
	e2 = 1/e1
	x = (e1 + e2) / 2
      END FUNCTION exflib_float_cosh_f

      FUNCTION exflib_float_tanh_f(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
	TYPE(exfloat) :: e1, e2
	e1 = exp(a)
        e2 = e1 * e1
	x = (e2 - 1) / (e2 + 1)
      END FUNCTION exflib_float_tanh_f

      FUNCTION exflib_float_asin_f(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
        INTEGER*8 :: prec, hint
        INTEGER*8 :: workmem(0:work_size)
        prec = c_prec
        hint = exflib_float_asin_FRT(x%num, a%num, prec, workmem)
        prec = c_prec
        CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_asin_f

      FUNCTION exflib_float_acos_f(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
        TYPE(exfloat) :: pi
        INTEGER*8 :: prec, hint
        INTEGER*8 :: workmem(0:work_size)
        prec = c_prec
        CALL exflib_float_ctoef_FRT(pi%num, 'pi' // ACHAR(0), prec);
        pi%num(0) = pi%num(0) - 1
        prec = c_prec
        hint = exflib_float_asin_FRT(x%num, a%num, prec, workmem)
        prec = c_prec
        hint = exflib_float_sub_FRT(x%num, pi%num, x%num, prec)
        prec = c_prec
        CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_acos_f

      FUNCTION exflib_float_atan_f(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
        INTEGER*8 :: prec, hint
        INTEGER*8 :: workmem(0:work_size)
        prec = c_prec
#ifdef _OPENMP
        !$OMP CRITICAL ( EXFLIB_INV_TABLE_ATAN )
        IF ( inv_table_atan(0) == 0 ) THEN

           CALL exflib_init_reciprocal_table_FRT( inv_table_atan, prec+2 )

        END IF
        !$OMP END CRITICAL ( EXFLIB_INV_TABLE_ATAN )
#endif
        hint = exflib_float_atan_FRT(x%num, a%num, prec, workmem, inv_table_atan)
        prec = c_prec
        CALL exflib_float_round_near_FRT(x%num, hint, prec)
      END FUNCTION exflib_float_atan_f

      ! integer and floating part
      FUNCTION exflib_float_mod_f(a, b) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a, b
        TYPE(exfloat) :: x
	INTEGER*8 :: prec
        INTEGER*8 :: workmem(0:work_size)
	prec = c_prec
        CALL exflib_float_fmod_FRT(x%num, a%num, b%num, prec, workmem)
      END FUNCTION exflib_float_mod_f

!      FUNCTION exflib_float_fraction_f(a) RESULT(x)
!        TYPE(exfloat), INTENT(IN) :: a
!        TYPE(exfloat) :: x
!	 INTEGER*8 :: prec
!        INTEGER*8 :: workmem(0:work_size)
! 	 prec = c_prec
!        CALL exflib_float_modf_FRT(a%num, workmem, x%num, prec)
!      END FUNCTION exflib_float_fraction_f

      FUNCTION exflib_float_int_f(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
	INTEGER*8 :: prec
        INTEGER*8 :: workmem(0:work_size)
	prec = c_prec
        CALL exflib_float_modf_FRT(a%num, x%num, workmem, prec)
      END FUNCTION exflib_float_int_f

      FUNCTION exflib_float_floor_f(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
	INTEGER*8 :: prec
        INTEGER*8 :: workmem(0:work_size)
	prec = c_prec
        CALL exflib_float_floor_FRT(x%num, a%num, prec, workmem)
      END FUNCTION exflib_float_floor_f

      FUNCTION exflib_float_ceiling_f(a) RESULT(x)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
	INTEGER*8 :: prec
        INTEGER*8 :: workmem(0:work_size)
	prec = c_prec
        CALL exflib_float_ceil_FRT(x%num, a%num, prec, workmem)
      END FUNCTION exflib_float_ceiling_f

      !--------------------------------------------------
      ! complex type
      !--------------------------------------------------
      ! complex type constructor
!      ! z = a
!      FUNCTION exflib_cmplx_z(a) RESULT(z)
!        TYPE(ezfloat), INTENT(IN) :: a
!        TYPE(ezfloat) :: z
!
!        NOT IMPLEMENT : to avoid ambiguity with 'double cmplx(ezfloat)'
!
!      END FUNCTION exflib_cmplx_z

      ! z = a + 0*i
      FUNCTION exflib_cmplx_e(a) RESULT(z)
        CHARACTER(*), INTENT(IN) :: a
        TYPE(ezfloat) :: z
        z%re = a
        z%im = 0
      END FUNCTION exflib_cmplx_e

      ! z = a + 0*i
      FUNCTION exflib_cmplx_m(a) RESULT(z)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(ezfloat) :: z
        z%re = a
        z%im = 0
      END FUNCTION exflib_cmplx_m

      ! z = a + 0*i
      FUNCTION exflib_cmplx_q(a) RESULT(z)
        INTEGER*8, INTENT(IN) :: a
        TYPE(ezfloat) :: z
        z%re = a
        z%im = 0
      END FUNCTION exflib_cmplx_q

      ! z = a + 0*i
      FUNCTION exflib_cmplx_i(a) RESULT(z)
        INTEGER*4, INTENT(IN) :: a
        TYPE(ezfloat) :: z
        z%re = a
        z%im = 0
      END FUNCTION exflib_cmplx_i

      ! z = a + b*i
      FUNCTION exflib_cmplx_ee(a, b) RESULT(z)
        CHARACTER(*), INTENT(IN) :: a, b
        TYPE(ezfloat) :: z
        z%re = a
        z%im = b
      END FUNCTION exflib_cmplx_ee

      ! z = a + b*i
      FUNCTION exflib_cmplx_em(a, b) RESULT(z)
        CHARACTER(*), INTENT(IN) :: a
        TYPE(exfloat) :: b
        TYPE(ezfloat) :: z
        z%re = a
        z%im = b
      END FUNCTION exflib_cmplx_em

      ! z = a + b*i
      FUNCTION exflib_cmplx_me(a, b) RESULT(z)
        TYPE(exfloat) :: a
        CHARACTER(*), INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a
        z%im = b
      END FUNCTION exflib_cmplx_me

      ! z = a + b*i
      FUNCTION exflib_cmplx_eq(a, b) RESULT(z)
        CHARACTER(*), INTENT(IN) :: a
        INTEGER*8 :: b
        TYPE(ezfloat) :: z
        z%re = a
        z%im = b
      END FUNCTION exflib_cmplx_eq

      ! z = a + b*i
      FUNCTION exflib_cmplx_qe(a, b) RESULT(z)
        INTEGER*8 :: a
        CHARACTER(*), INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a
        z%im = b
      END FUNCTION exflib_cmplx_qe

      ! z = a + b*i
      FUNCTION exflib_cmplx_ei(a, b) RESULT(z)
        CHARACTER(*), INTENT(IN) :: a
        INTEGER*4 :: b
        TYPE(ezfloat) :: z
        z%re = a
        z%im = b
      END FUNCTION exflib_cmplx_ei

      ! z = a + b*i
      FUNCTION exflib_cmplx_ie(a, b) RESULT(z)
        INTEGER*4 :: a
        CHARACTER(*), INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a
        z%im = b
      END FUNCTION exflib_cmplx_ie

      ! z = a + b*i
      FUNCTION exflib_cmplx_mm(a, b) RESULT(z)
        TYPE(exfloat), INTENT(IN) :: a, b
        TYPE(ezfloat) :: z
        z%re = a
        z%im = b
      END FUNCTION exflib_cmplx_mm

      ! z = a + b*i
      FUNCTION exflib_cmplx_mq(a, b) RESULT(z)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a
        z%im = b
      END FUNCTION exflib_cmplx_mq

      ! z = a + b*i
      FUNCTION exflib_cmplx_qm(a, b) RESULT(z)
        INTEGER*8, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a
        z%im = b
      END FUNCTION exflib_cmplx_qm

      ! z = a + b*i
      FUNCTION exflib_cmplx_mi(a, b) RESULT(z)
        TYPE(exfloat), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a
        z%im = b
      END FUNCTION exflib_cmplx_mi

      ! z = a + b*i
      FUNCTION exflib_cmplx_im(a, b) RESULT(z)
        INTEGER*4, INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a
        z%im = b
      END FUNCTION exflib_cmplx_im

      ! z = a + b*i
      FUNCTION exflib_cmplx_qq(a, b) RESULT(z)
        INTEGER*8, INTENT(IN) :: a, b
        TYPE(ezfloat) :: z
        z%re = a
        z%im = b
      END FUNCTION exflib_cmplx_qq

      ! z = a + b*i
      FUNCTION exflib_cmplx_qi(a, b) RESULT(z)
        INTEGER*8, INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a
        z%im = b
      END FUNCTION exflib_cmplx_qi

      ! z = a + b*i
      FUNCTION exflib_cmplx_iq(a, b) RESULT(z)
        INTEGER*4, INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(ezfloat) :: z
        z%re = a
        z%im = b
      END FUNCTION exflib_cmplx_iq

      ! z = a + b*i
      FUNCTION exflib_cmplx_ii(a, b) RESULT(z)
        INTEGER*4, INTENT(IN) :: a, b
        TYPE(ezfloat) :: z
        z%re = a
        z%im = b
      END FUNCTION exflib_cmplx_ii

      ! ezfloat to COMPLEX*16 (double, double)
      FUNCTION exflib_cmplx_zw(z) RESULT(d)
        TYPE(ezfloat), INTENT(IN) :: z
        COMPLEX*16 :: d
        d = DCMPLX( DBLE(z%re), DBLE(z%im) )
      END FUNCTION exflib_cmplx_zw

      ! constructor in polar form
      FUNCTION exflib_polar_ee(r, t) RESULT(z)
        CHARACTER(*), INTENT(IN) :: r, t
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: tmp1, tmp2
        tmp1 = r
        tmp2 = t
        z%re = tmp1 * cos(tmp2)
        z%im = tmp1 * sin(tmp2)
      END FUNCTION exflib_polar_ee

      FUNCTION exflib_polar_em(r, t) RESULT(z)
        CHARACTER(*), INTENT(IN) :: r
        TYPE(exfloat), INTENT(IN) :: t
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: tmp
        tmp = r
        z%re = tmp * cos(t)
        z%im = tmp * sin(t)
      END FUNCTION exflib_polar_em

      FUNCTION exflib_polar_me(r, t) RESULT(z)
        TYPE(exfloat), INTENT(IN) :: r
        CHARACTER(*), INTENT(IN) :: t
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: tmp
        tmp = t
        z%re = r * cos(tmp)
        z%im = r * sin(tmp)
      END FUNCTION exflib_polar_me

      FUNCTION exflib_polar_eq(r, t) RESULT(z)
        CHARACTER(*), INTENT(IN) :: r
        INTEGER*8, INTENT(IN) :: t
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: tmp1, tmp2
        tmp1 = r
        tmp2 = t
        z%re = tmp1 * cos(tmp2)
        z%im = tmp1 * sin(tmp2)
      END FUNCTION exflib_polar_eq

      FUNCTION exflib_polar_qe(r, t) RESULT(z)
        INTEGER*8, INTENT(IN) :: r
        CHARACTER(*), INTENT(IN) :: t
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: tmp
        tmp = t
        z%re = r * cos(tmp)
        z%im = r * sin(tmp)
      END FUNCTION exflib_polar_qe

      FUNCTION exflib_polar_ei(r, t) RESULT(z)
        CHARACTER(*), INTENT(IN) :: r
        INTEGER*4, INTENT(IN) :: t
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: tmp1, tmp2
        tmp1 = r
        tmp2 = t
        z%re = tmp1 * cos(tmp2)
        z%im = tmp1 * sin(tmp2)
      END FUNCTION exflib_polar_ei

      FUNCTION exflib_polar_ie(r, t) RESULT(z)
        INTEGER*4, INTENT(IN) :: r
        CHARACTER(*), INTENT(IN) :: t
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: tmp
        tmp = t
        z%re = r * cos(tmp)
        z%im = r * sin(tmp)
      END FUNCTION exflib_polar_ie

      FUNCTION exflib_polar_mm(r, t) RESULT(z)
        TYPE(exfloat), INTENT(IN) :: r, t
        TYPE(ezfloat) :: z
        z%re = r * cos(t)
        z%im = r * sin(t)
      END FUNCTION exflib_polar_mm

      FUNCTION exflib_polar_mq(r, t) RESULT(z)
        TYPE(exfloat), INTENT(IN) :: r
        INTEGER*8, INTENT(IN) :: t
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: tmp
        tmp = t
        z%re = r * cos(tmp)
        z%im = r * sin(tmp)
      END FUNCTION exflib_polar_mq

      FUNCTION exflib_polar_qm(r, t) RESULT(z)
        INTEGER*8, INTENT(IN) :: r
        TYPE(exfloat), INTENT(IN) :: t
        TYPE(ezfloat) :: z
        z%re = r * cos(t)
        z%im = r * sin(t)
      END FUNCTION exflib_polar_qm

      FUNCTION exflib_polar_mi(r, t) RESULT(z)
        TYPE(exfloat), INTENT(IN) :: r
        INTEGER*4, INTENT(IN) :: t
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: tmp
        tmp = t
        z%re = r * cos(tmp)
        z%im = r * sin(tmp)
      END FUNCTION exflib_polar_mi

      FUNCTION exflib_polar_im(r, t) RESULT(z)
        INTEGER*4, INTENT(IN) :: r
        TYPE(exfloat), INTENT(IN) :: t
        TYPE(ezfloat) :: z
        z%re = r * cos(t)
        z%im = r * sin(t)
      END FUNCTION exflib_polar_im

      FUNCTION exflib_polar_qq(r, t) RESULT(z)
        INTEGER*8, INTENT(IN) :: r, t
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: tmp
        tmp = t
        z%re = r * cos(tmp)
        z%im = r * sin(tmp)
      END FUNCTION exflib_polar_qq

      FUNCTION exflib_polar_qi(r, t) RESULT(z)
        INTEGER*8, INTENT(IN) :: r
        INTEGER*4, INTENT(IN) :: t
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: tmp
        tmp = t
        z%re = r * cos(tmp)
        z%im = r * sin(tmp)
      END FUNCTION exflib_polar_qi

      FUNCTION exflib_polar_iq(r, t) RESULT(z)
        INTEGER*4, INTENT(IN) :: r
        INTEGER*8, INTENT(IN) :: t
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: tmp
        tmp = t
        z%re = r * cos(tmp)
        z%im = r * sin(tmp)
      END FUNCTION exflib_polar_iq

      FUNCTION exflib_polar_ii(r, t) RESULT(z)
        INTEGER*4, INTENT(IN) :: r
        INTEGER*4, INTENT(IN) :: t
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: tmp
        tmp = t
        z%re = r * cos(tmp)
        z%im = r * sin(tmp)
      END FUNCTION exflib_polar_ii

      ! real part
      FUNCTION exflib_ezfloat_real(a) RESULT(x)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
        x = a%re
      END FUNCTION exflib_ezfloat_real

      ! imaginary part
      FUNCTION exflib_ezfloat_imag(a) RESULT(x)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
        x = a%im
      END FUNCTION exflib_ezfloat_imag

      FUNCTION exflib_ezfloat_aimag(a) RESULT(x)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
        x = a%im
      END FUNCTION exflib_ezfloat_aimag

      ! complex conjugate
      FUNCTION exflib_ezfloat_conjg(a) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat) :: z
        z%re =  a%re
        z%im = -a%im
      END FUNCTION exflib_ezfloat_conjg

      ! argument
      FUNCTION exflib_ezfloat_arg(a) RESULT(x)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
        TYPE(exfloat) :: PI
        INTEGER*8 :: prec

        prec = c_prec
        CALL exflib_float_ctoef_FRT(PI%num, 'pi' // ACHAR(0), prec);

        IF ( a%re == 0 ) THEN

           x = PI / 2
           IF ( a%im < 0 ) THEN
              x = -x
           ELSE IF( a%im == 0 ) THEN
              x = 0                       ! Arg(0) is undefined
           END IF

        ELSE IF( a%re > 0 ) THEN

           x = atan( a%im / a%re )

        ELSE

           IF ( a%im < 0 ) THEN
              x = - PI + atan( a%im / a%re )
           ELSE
              x = PI - atan( - a%im / a%re)
           END IF

        END IF

      END FUNCTION exflib_ezfloat_arg

      !--------------------------------------------------
      ! built-in functions for complex variable
      !--------------------------------------------------
      FUNCTION exflib_float_abs_z(a) RESULT(x)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat) :: x
        TYPE(exfloat) :: s, r, i

        IF ( abs(a%re) > abs(a%im) ) THEN
           s = abs(a%re)
        ELSE
           s = abs(a%im)
        END IF

        IF ( s == 0 ) THEN
           x = 0
           RETURN
        END IF

        r = a%re / s
        i = a%im / s
        x = sqrt(r*r + i*i) * s

      END FUNCTION exflib_float_abs_z

      FUNCTION exflib_float_sqrt_z(a) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: t, u

        IF ( a%re == 0 ) THEN

           z%re = sqrt( abs( a%im ) / 2 )

           IF( a%im < 0 ) THEN
              z%im = - z%re
           ELSE
              z%im =   z%re
           END IF

        ELSE

           t = sqrt( 2 * ( abs(a) + abs( a%re ) ) )
           u = t / 2

           IF ( a%re > 0 ) THEN

              z%re = u
              z%im = a%im / t

           ELSE

              z%re = abs(a%im) / t

              IF ( a%im < 0 ) THEN
                 z%im = -u
              ELSE
                 z%im =  u
              END IF

           END IF

        END IF

      END FUNCTION exflib_float_sqrt_z

!      FUNCTION exflib_float_cubic_z(a) RESULT(z)
!        TYPE(ezfloat), INTENT(IN) :: a
!        TYPE(ezfloat) :: z
!
!        TODO
!
!      END FUNCTION exflib_float_cubi_z

      FUNCTION exflib_float_sin_z(a) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: e1, e2

        ! sin(x+iy) = cmplx( sin(x)*cosh(y), cos(x)*sinh(y) )
        e1 = exp(a%im)
        e2 = 1 / e1
        z%re = sin(a%re) * (e1 + e2) / 2
        z%im = cos(a%re) * (e1 - e2) / 2
      END FUNCTION exflib_float_sin_z

      FUNCTION exflib_float_cos_z(a) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: e1, e2

        ! cos(x+iy) = cmplx( cos(x)*cosh(y), -sin(x)*sinh(y) )
        e1 = exp(a%im)
        e2 = 1 / e1
        z%re =   cos(a%re) * (e1 + e2) / 2
        z%im = - sin(a%re) * (e1 - e2) / 2
      END FUNCTION exflib_float_cos_z

      FUNCTION exflib_float_tan_z(a) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat) :: z
        z = sin(a) / cos(a)
      END FUNCTION exflib_float_tan_z

      FUNCTION exflib_float_exp_z(a) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: r

        ! exp(x+iy) = exp(x) * exp(iy) = exp(x) * ( cos(y) + i sin(y) )
        r = exp(a%re)
        z%re = r * cos(a%im)
        z%im = r * sin(a%im)
      END FUNCTION exflib_float_exp_z

      FUNCTION exflib_float_pow_zz(a, b) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a, b
        TYPE(ezfloat) :: z

        IF ( a == 0 ) THEN
           IF ( b == 0 ) THEN
              z = 1             ! 0^0 should be NaN
           ELSE
              z = 0
           END IF
        ELSE
           z = exp( log( a ) * b )
        END IF

      END FUNCTION exflib_float_pow_zz

      FUNCTION exflib_float_pow_zm(a, b) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(exfloat), INTENT(IN) :: b
        TYPE(ezfloat) :: z

        IF ( a == 0 ) THEN
           IF ( b == 0 ) THEN
              z = 1             ! 0^0 should be NaN
           ELSE
              z = 0
           END IF
        ELSE
           z = exp( log( a ) * b )
        END IF

      END FUNCTION exflib_float_pow_zm

      FUNCTION exflib_float_pow_mz(a, b) RESULT(z)
        TYPE(exfloat), INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat) :: z

        IF ( a > 0 ) THEN
           z = polar( pow(a, b%re), log(a) * b%im )
        ELSE
           z = pow( CMPLX(a, 0), b )
        END IF

      END FUNCTION exflib_float_pow_mz

      FUNCTION exflib_float_pow_zq(a, b) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*8, INTENT(IN) :: b
        TYPE(ezfloat) :: z
        TYPE(ezfloat) :: tmp

        INTEGER*8 :: n
        n = b
        IF ( n < 0 ) THEN
           n = - n
        END IF

        z = 1               ! 0^0 should be NaN
        tmp = a
        DO WHILE ( n > 0 )

           IF ( (n - 2*(n/2)) == 1 ) THEN        ! MOD(n,2)
              z = z * tmp
           END IF

           n = n / 2
           tmp = tmp * tmp

        END DO

        IF ( b < 0 ) THEN
           z = 1 / z
        END IF
      END FUNCTION exflib_float_pow_zq

      FUNCTION exflib_float_pow_qz(a, b) RESULT(z)
        INTEGER*8, INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: tmp
        tmp = a
        z = exflib_float_pow_mz(tmp, b)
      END FUNCTION exflib_float_pow_qz

      FUNCTION exflib_float_pow_zi(a, b) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        INTEGER*4, INTENT(IN) :: b
        TYPE(ezfloat) :: z
        TYPE(ezfloat) :: tmp

        INTEGER*4 :: n
        n = b
        IF ( n < 0 ) THEN
           n = - n
        END IF

        z = 1               ! 0^0 should be NaN
        tmp = a
        DO WHILE ( n > 0 )

           IF ( (n - 2*(n/2)) == 1 ) THEN        ! MOD(n,2)
              z = z * tmp
           END IF

           n = n / 2
           tmp = tmp * tmp

        END DO

        IF ( b < 0 ) THEN
           z = 1 / z
        END IF

      END FUNCTION exflib_float_pow_zi

      FUNCTION exflib_float_pow_iz(a, b) RESULT(z)
        INTEGER*4, INTENT(IN) :: a
        TYPE(ezfloat), INTENT(IN) :: b
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: tmp
        tmp = a
        z = exflib_float_pow_mz(tmp, b)
      END FUNCTION exflib_float_pow_iz

      FUNCTION exflib_float_log_z(a) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat) :: z

        ! log(x+iy) = cmplx( log( abs(x+iy) ), arg(x+iy) )
        z%re = log( abs(a) )
        z%im = arg(a)
      END FUNCTION exflib_float_log_z

      FUNCTION exflib_float_log10_z(a) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat) :: z
	TYPE(exfloat) :: ln10, invln10
	INTEGER*8 :: prec
	prec = c_prec
	CALL exflib_float_ctoef_FRT(ln10%num, 'ln10' // ACHAR(0), prec)
        invln10 = 1 / ln10
        z%re = log( abs(a) ) * invln10
        z%im = arg(a) * invln10
      END FUNCTION exflib_float_log10_z

      FUNCTION exflib_float_sinh_z(a) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: e1, e2

        ! sinh(x+iy) = cmplx( sinh(x)*cos(y), cosh(x)*sin(y) )
        e1 = exp(a%re)
        e2 = 1 / e1
        z%re = (e1 - e2) / 2 * cos(a%im)
        z%im = (e1 + e2) / 2 * sin(a%im)
      END FUNCTION exflib_float_sinh_z

      FUNCTION exflib_float_cosh_z(a) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat) :: z
        TYPE(exfloat) :: e1, e2

        ! cosh(x+iy) = cmplx( cosh(x)*cos(y), sinh(x)*sin(y) )
        e1 = exp(a%re)
        e2 = 1 / e1
        z%re = (e1 + e2) / 2 * cos(a%im)
        z%im = (e1 - e2) / 2 * sin(a%im)
      END FUNCTION exflib_float_cosh_z

      FUNCTION exflib_float_tanh_z(a) RESULT(z)
        TYPE(ezfloat), INTENT(IN) :: a
        TYPE(ezfloat) :: z
        z = sinh(a) / cosh(a)
      END FUNCTION exflib_float_tanh_z

   END MODULE exflib

!----------------------------------------------------------------------
! End of file
!----------------------------------------------------------------------
